#define LOGFILE		"dtbentry.log"
#define C_BUF_SIZE	2048
#define L_BUF_SIZE	2048
#define	REP_BUF_S_SIZE	1024

#include <errno.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <netdb.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <resolv.h>
#include <net/if.h>
#include <sys/socket.h>
#include <openssl/ssl.h>
#include <arpa/inet.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include "config.h"
#include "types.h"


char		*site_ip[ MAX_CONNECTIONS ];
unsigned long	bouncer_pass;
key_t		key;
int		target_hosts = 0;
int		current_target = 0;
int		read_ident = -1;
int		send_ident = -1;
int		client_port = -1;
int		site_port[ MAX_CONNECTIONS ];
int		pipes[ 2 ];
int		distr_mode = -1;
int		bouncer_port = -1;
int		client_num = 0;

struct	entry_container *container;
extern	char		*rfc931();
extern	int		errno;



/*
 * void log( char *format, ...);
 *
 * Writes stuff to LOGFILE
 *
 */
void
log( char *fmt, ...) {
 time_t         timenow;
 FILE           *file;
 va_list        ap;

 va_start(ap, fmt); 
 timenow = time(NULL);

 if ((file = fopen(LOGFILE, "a+")) != NULL ) {
	fprintf( file, "%.24s - ", ctime( &timenow ) );
	vfprintf( file, fmt, ap );
	fclose( file );
	}
}




/*
 * void init_container() 
 *
 * Initilizes shared memory for us to use
 *
 */
void init_container() {
 int	shmid;
 int	n;

 if ((shmid = shmget( (key_t)key, (int)(sizeof(struct entry_container) * MAX_CONNECTIONS), IPC_CREAT | 0644 )) == -1) {
	printf("(SHMGET) Failed!\n");
	return;
	}

 if ((container = (struct entry_container *)shmat( shmid, NULL, 0)) == (struct entry_container *)-1) {
	printf("(SHMAT) Failed!\n");
	return;
	}

	/* Mark container as destroyed - so it will be cleared when we die */

 shmctl(shmid, IPC_RMID, 0);

 container[0].storage = 0;

 for ( n = 0 ; n < MAX_CONNECTIONS ; n++ ) { 
	container[n].user_online = 0;
	container[n].bnc_status = OFFLINE;
	}
}




/*
* int add_container( int fd, char *ip ) 
*
* Adds new container in to reserver area in shared memory
*
*/
int add_container( int fd, char *ip ) {
 struct	entry_container	ptr;
 int			n;

 bzero(&ptr, sizeof(struct entry_container));

 strncpy(ptr.bnc_ip, ip, 16);
 ptr.bnc_status = WAIT_AUTH;
 ptr.bnc_fd = fd;

	/* Seek for empty area in mem */

 for ( n = 0 ; n < MAX_CONNECTIONS ; n++ ) {
	if ( container[ n ].bnc_status != OFFLINE ) continue;
	ptr.user_online = container[ n ].user_online;
	ptr.storage = container[n].storage;
	ptr.bnc_reply = container[n].bnc_reply;
	memcpy(&container[n], &ptr, sizeof(struct entry_container));
	return n;
	}

	/* No free areas - return error */

 return -1;
}




/*
* int pipe_send_command( int bnc_n, unsigned char cmd, int owner, char *host, int port )
*
* Sends command to chosen bnc through pipe and verifies that connection 
* to bouncer is still alive.
*/
int
pipe_send_command( int bnc_n, unsigned char cmd, int owner, char *host, int port ) {
 struct	cmd_t		command;

 if (container[bnc_n].bnc_status == OFFLINE) {
	return -1;
	}

 container[owner].bnc_reply = -1;

 command.target = bnc_n;
 command.cmd = cmd;
 command.owner = owner;

 memcpy(command.hostname, host, 16);
 command.port = port;
 write(pipes[1], &command, sizeof(struct cmd_t));

 return 0;
}



/*
* void direct_send_command( int bnc_n, unsigned char cmd, int owner, char *host, int port )
*
* Sends command directly to chosen bnc
*
*/
void
direct_send_command( int bnc_n, unsigned char cmd, int owner, char *host, int port ) {
 struct	cmd		command;

 command.cmd = cmd;
 command.owner = owner;
 memcpy( command.hostname, host, 16 );
 command.port = port;
 write(container[bnc_n].bnc_fd, &command, sizeof(struct cmd));
}




/*
* unsigned long get_bnc_reply( int owner )
*
* Waits until reply from bouncer arrives
* If no answer in 20secs, then abort with value 0.
*
*/
unsigned long
get_bnc_reply( int owner ) {
 unsigned long	n = 0;
 unsigned 	wait = 0;

 while ( container[owner].bnc_reply == -1 && wait < 5000) {
	usleep(50000);
	wait++;
	}

 n = container[owner].bnc_reply;
 container[owner].bnc_reply = -1;

 return n;
}





/*
* int select_bnc()
*
* Select trafficbouncer from container array with chosen method.
*
*/
int
select_bnc() {
 struct	timeval	tv;
 unsigned char	n = 0;
 unsigned char	bncs = 0;
 unsigned char	bnc[MAX_CONNECTIONS];
 unsigned char	ROUND_ROBIN = container[0].storage;
 unsigned long	min = 0;

 switch	( distr_mode ) {
	case 0:
		 bncs = 1;
		 while ( ! container[ ROUND_ROBIN ].bnc_status == ACTIVE && n < MAX_CONNECTIONS ) {
			ROUND_ROBIN++;
			if ( ROUND_ROBIN >= MAX_CONNECTIONS ) ROUND_ROBIN = 0;
			n++;
			}
		 bnc[0] = ROUND_ROBIN;
		 container[0].storage = ROUND_ROBIN + 1;
		break;
	case 1:
		 for ( ; n < MAX_CONNECTIONS ; n++ ) if ((container[n].bnc_connections <= min || bncs == 0 ) && container[n].bnc_status == ACTIVE ) {
			if ( container[n].bnc_connections != min ) {
				min = container[n].bnc_connections;
				bncs = 0;
				}
			bnc[bncs] = n;
			bncs++;
			}
		break;
	case 2:
		 for ( ; n < MAX_CONNECTIONS ; n++ ) if ((container[n].bnc_bytes <= min || bncs == 0 ) && container[n].bnc_status == ACTIVE) {
			 if ( container[n].bnc_bytes != min ) {
				 min = container[ n ].bnc_bytes;
				 bncs = 0;
				}
			bnc[ bncs ] = n;
			bncs++;
			}
		break;
	 case 3:
		 for ( ; n < MAX_CONNECTIONS ; n++ ) if ( container[ n ].bnc_status == ACTIVE ) {
			bnc[ bncs ] = n;
			bncs++;
			}
	}

	/* Return error, if no active bouncers */


 if ( container[ bnc[ 0 ] ].bnc_status != ACTIVE ) {
	return -1;
	}

	/* We will select random bouncer from list, if multple bouncers were chosen */

 if ( bncs > 1 ) {
	gettimeofday( &tv, NULL );
	srandom( tv.tv_usec );
	n = (int)(bncs * 1. * rand() / ( RAND_MAX + 1.0 ));
	bnc[0] = bnc[n];
	}

 return bnc[0];
}





/*
 * int modify_ftp_response( char *s, int *s_size, int *entrynum )
 *
 * Modifies ftp's responses (ie pasv)
 *
 *
 */
int
modify_ftp_response( char *s, int *s_size, int user, int *entrynum ) {
 char	*s_p,
	*ipport[6],
	ip[16],
	pasvreply[] = { '(' , ',' , ',' , ',' , ',' , ',' , ')' };
 int	n;
 unsigned 	port;
 unsigned char	*c_port;

	/* 227 is reserved for PASV reply */

 if ( ! strncmp( s, "227 ", 4 ) ) {
	 s_p = s;
	 n = *entrynum;
	 pipe_send_command( n, REQ_CLEAR, user, NULLSTR, 0 );
	 for ( n = 0 ; n < 6 ; n++ ) {
		for ( ; *s_p != pasvreply[n] && *s_p != '\n' ; s_p++ );
		if ( *s_p != pasvreply[n] ) return OTHER;
		*s_p = 0;
		s_p++;
		ipport[n] = s_p;
		}
	 port = atoi(ipport[4]) * 256 + atoi(ipport[5]);
	 sprintf(ip, "%s.%s.%s.%s", ipport[0], ipport[1], ipport[2], ipport[3]);

	 if ( (n = select_bnc()) == - 1 ) {
		*s_size = sprintf(s, "500 No active trafficbouncers.\r\n" );
		return OTHER;
		}

	 *entrynum = n;
	 log("Sending PASV request to bnc [%i]\n", n);
	 if (pipe_send_command( n, REQ_PASV, user, ip, port ) != 0) {
		*s_size = sprintf(s, "500 Connection to trafficbouncer lost.\r\n" );
		return OTHER;
		}

	 if (( port = get_bnc_reply(user)) == 0) {
		*s_size = sprintf(s, "500 No response from traffic bouncer [%s].\r\n", container[n].bnc_ip);
		return OTHER;
		}

	c_port = (unsigned char *)&port;

	strcpy(ip, container[n].bnc_ip);
	for ( n = 0 ; n < 16 ; n++ ) {
		if ( ip[n] == '.' ) ip[n] = ',';
		}

	*s_size = sprintf(s, "227 Entering Passive Mode (%s,%u,%u)\r\n", ip, *(c_port + 1), *c_port);
	return PASV;
	}

	/* 150 begins the transfer */

 if ( ! strncmp( s, "150 ", 4 ) ) return TRANSFER;

 return OTHER;
}




char nicetime[20];
char *nice_time(unsigned long secs) {
 int	hours = 0;
 int	mins = 0;


 while ( secs >= 3600 ) {
	hours++;
	secs -= 3600;
	}
 while ( secs >= 60 ) {
	mins++;
	secs -= 60;
	}

 if ( mins > 0 ) {
	snprintf(nicetime, 20, "%ih %im", hours, mins);
	} else {
	snprintf(nicetime, 20, "soon");
	}

 return nicetime;
}




/*
* int modify_client_command( char *s, int *s_size, int user, int *entrynum )
*
* Checks wheter we like command or not.
*
*/
int modify_client_command(char *s, int *s_size, int user, int *entrynum) {
 char			*s_p,
			*ipport[ 6 ],
			ip[ 16 ],
			portreply[] = { ' ' , ',' , ',' , ',' , ',' , ',' , '\r' },
			SuparSecret[24],
			*status[] = { "ACTIVE", "CLOSED", "AUTH" };
 int			n, l;
 int			port;
 int			bncs = 0;
 unsigned char		*c_port;
 unsigned long		bytes = 0;
 unsigned long		left = 0;

	/* Rewrite PORT to fit our purposes */

 if ( ! strncasecmp( "PORT ", s, 5 ) ) {
	 s_p = s;
	 pipe_send_command( *entrynum, REQ_CLEAR, user, NULLSTR, 0 );
	 for ( n = 0 ; n < 6 ; n++ ) {
		 for ( ; *s_p != portreply[n] && *s_p != '\n' ; s_p++ );
		 if ( *s_p != portreply[n] ) return 1;
		 *s_p = 0;
		 s_p++;
		 ipport[n] = s_p;
		}
	 port = atoi( ipport[ 4 ] ) * 256 + atoi( ipport[ 5 ] );
	 sprintf( ip, "%s.%s.%s.%s", ipport[0], ipport[1], ipport[2], ipport[3] );

	 if ((n = select_bnc()) == -1) {
		*s_size = sprintf(s, "500 No active trafficbouncers.\r\n" );
		return 0;
		}

	 *entrynum = n;
	 if ( pipe_send_command( n, REQ_PORT, user, ip, port ) != 0 ) {
		*s_size = sprintf(s, "500 Connection to trafficbouncer lost.\r\n" );
		return 0;
		}

	 if ((port = get_bnc_reply(user)) == 0) {
		*s_size = sprintf(s, "500 No response from traffic bouncer.\r\n" );
		return 0;
		}

	c_port = (unsigned char *)&port;

	strcpy( ip, container[ n ].bnc_ip );
	for ( n = 0 ; n < 16 ; n++ ) {
		if ( ip[ n ] == '.' ) ip[ n ] = ',';
		}

	*s_size = sprintf(s, "PORT %s,%i,%i\r\n", ip, *(c_port + 1), *c_port);
	return 1;
	}

	/* Our SuparSecret thingie */

 n = snprintf( SuparSecret, 24, "DTBNC INFO %u\r\n", bouncer_pass );

 if ( ! strncasecmp( SuparSecret, s, n ) ) {
	l = sprintf( s, "200- dTbnc stats:\r\n200-\r\n200- STATUS      IP                 TOTAL/LEFT     AVG SPEED    RESET\r\n200- -----------------------------------------------------------------\r\n" );
	for ( n = 0 ; n < MAX_CONNECTIONS ; n++ ) if ( container[ n ].bnc_status >= ACTIVE ) {
		bncs++;
		bytes += container[n].bnc_bytes;
		left += container[n].bnc_bytes_left;
		l += sprintf(s + l, "200- %-6s  %-16s %11.1f/%-8.1f %7.1fKB/s %s\r\n", status[ container[ n ].bnc_status - 1 ], container[ n ].bnc_ip, container[ n ].bnc_bytes / 1024., container[ n ].bnc_bytes_left / 1024., container[ n ].bnc_speed / 1024., nice_time( container[ n ].bnc_reset - time( NULL ) ) );
		}
	l += sprintf( s + l, "200- -----------------------------------------------------------------\r\n200- TOTAL   %-4i             %11.1f/%-8.1f\r\n200-\r\n200 End of list\r\n", bncs, bytes / 1024., left / 1024. ); 
	*s_size = l;
	return 0;
	}

	/* Do not allow client to begin transfer, if there is no traffic bouncer active */

 for ( n = 0 ; n < MAX_CONNECTIONS ; n++ ) if ( container[ n ].bnc_status == ACTIVE ) bncs++;

 if ( bncs == 0 ) {
	if ( *s_size == 6 ) {
		if ( ! memcmp( "LIST\r\n", s, 6 ) ||
		      ! memcmp( "RETR\r\n", s, 6 ) || 
		      ! memcmp( "APPE\r\n", s, 6 ) ||
		      ! memcmp( "STOR\r\n", s, 6 ) ) {
			*s_size = sprintf(s, "500 Cannot transfer. No link to trafficbouncer.\r\n", s );
			return 0;
			}
		} else if ( *s_size > 7 ) {
		if ( ! memcmp( "LIST ", s, 5 ) ||
		      ! memcmp( "RETR ", s, 5 ) ||
		      ! memcmp( "APPE ", s, 5 ) ||
		      ! memcmp( "STOR ", s, 5 ) ) {
			*s_size = sprintf(s, "500 Cannot transfer. No link to trafficbouncer.\r\n", s );
			return 0;
			}
		}
	}

 return 1;
}




/*
* int tls_login( int client, int server )
*
* Performs tls_login on server.
*
*
*/
int
tls_login( int client, int server, char *host ) {
#ifdef TLS
 char	l_buf[ L_BUF_SIZE ],
	t_buf[ L_BUF_SIZE ];
 int	b,
	t_b;

 b = read( server, &l_buf, L_BUF_SIZE );
 write( client, &l_buf, b );

 t_b = read( client, &t_buf, L_BUF_SIZE );
 while ( memcmp( t_buf, "AUTH TLS", 8 ) == 0 ) {
	 write( client, "500 TLS not supported.\r\n", 25 );
	 t_b = read( client, &t_buf, L_BUF_SIZE );
	}

 tls_hostname = host;
 if ( tls_init() ) {
	return -1;
	}

 write(server, "AUTH TLS\r\n", 10);
 read(server, &l_buf, L_BUF_SIZE);

 if ( ! strncasecmp("234 AUTH TLS successful\r\n", l_buf, 25) ) {
	 if ( tls_connect_ctrl( server ) ) {
		 return -1;
		}
        }
 WRITE( server, &t_buf, t_b );

#endif
 return 0;
}




/*
* void ident_call( int client, int server, char *remote_host, struct sockaddr_in *client_addr, struct sockaddr_in *entry_addr )
*
* Reads and forwards ident.
*
*/
void
ident_call( int client, int server, char *remote_host, struct sockaddr_in *client_addr, struct sockaddr_in *entry_addr ) {
 char	*ident,
	*i_buf = NULL,
	*remote_ip;
 int	n;

 if ( read_ident == 1 ) {
	ident = rfc931( client_addr, entry_addr );
	remote_ip = inet_ntoa( client_addr->sin_addr );
	i_buf = malloc( strlen( ident ) + strlen( remote_ip ) + strlen( remote_host ) + 9 );
	n = sprintf( i_buf, "IDNT %s@%s:%s\r\n", ident, remote_ip, remote_host );
 	} else if (read_ident == 2) {
	i_buf = malloc(2048);
	n = read(client, i_buf, 2048);
	}

 if ( i_buf != NULL ) {
	if (send_ident == 1) write(server, i_buf, n);
	free(i_buf);
	}
}





/*
* void control_loop( int client, int server, int c_num ) 
*
* Transfers ftp control data from client to server and server to client
*
*/
void
control_loop( int client, int server, int c_num ) {
 fd_set		read_ds;
 int		n = 0, b_fd,
		REP_BUF_S = 0,
		REP_BUF_SIZE = REP_BUF_S_SIZE;
 char		c_buf[ C_BUF_SIZE ],
		*REP_BUF;
 int		BNC = 0;


 REP_BUF = malloc( REP_BUF_SIZE );

 b_fd = client > server ? client : server;
 b_fd++;

	/* Start listening for client and server */

 while ( 1 ) {
	FD_ZERO(&read_ds);
	FD_SET(server, &read_ds);
	FD_SET(client, &read_ds);

		/* Server has sent data */

	select(b_fd, &read_ds, 0, 0, 0);

	if (FD_ISSET(server, &read_ds)) {
		n = READ(server, &c_buf, C_BUF_SIZE);
		if (n <= 0) break;

		if (n + REP_BUF_S >= REP_BUF_SIZE) {
			REP_BUF_SIZE += n + 10;
			REP_BUF = realloc(REP_BUF, REP_BUF_SIZE);
			}

		if (memcmp( c_buf + n - 2, "\r\n", 2 ) == 0) {
			memcpy( REP_BUF + REP_BUF_S, c_buf, n );
			REP_BUF_S += n;
			REP_BUF[ REP_BUF_S ] = 0;
			switch ( modify_ftp_response( REP_BUF, &REP_BUF_S, c_num, &BNC ) ) {
				case OTHER:
				case PASV:
					if (WRITE(client, REP_BUF, REP_BUF_S) != REP_BUF_S) {
						goto end;
						}
					break;
				case TRANSFER:
					if (WRITE(client, REP_BUF, REP_BUF_S) != REP_BUF_S) {
						goto end;
						}
					pipe_send_command(BNC, REQ_ACCEPT, c_num, NULLSTR, 0);
					log( "Requested traffic bouncer [%s] to accept transfer\n", container[ BNC ].bnc_ip );
					break;
				}
			*REP_BUF = REP_BUF_S = 0;
			} else {
			strncpy(REP_BUF + REP_BUF_S, c_buf, n);
			REP_BUF_S += n;
			}
		}

		/* Client has sent data */

	 if (FD_ISSET(client, &read_ds)) {
		n = READ(client, &c_buf, C_BUF_SIZE);
		if (n <= 0) break;
		if (modify_client_command(c_buf, &n, c_num, &BNC)) {
			if (WRITE(server, &c_buf, n) != n) {
				goto end;
				}
			} else {
			if (WRITE(client, &c_buf, n) != n) {
				goto end;
				}
			}
		}
 	}

end:
	/* Make sure that client doesnt leave anything open on tb */

 pipe_send_command( BNC, REQ_CLEAR, c_num, NULLSTR, 0 );
 free( REP_BUF );
}







/*
* void read_config( char *config )
*
* Reads dynamic variables from config file and 
* allocates memory for them
*
*/
void
read_config( char *config ) {
 FILE	*cf;
 int	c_size,
	n, m, l;
 char	*c_buf,
	*t,
 	host[ MAXHOSTNAMELEN ],
	port[6],
	pass[24],
	readident[5],
	sendident[4],
	t_key[12],
	clientport[6],
	bouncerport[6],
	distrmode[16];
 int	i_start = 0,
	v_start = 0;

	/* Open Config file and buffer it */

 if ( ( cf = fopen( config, "r" ) ) == NULL ) {
	 printf( "dT-entry encountered fatal failure - config not found!\n");
	 exit( 2 );
	}

 fseek( cf, 0, SEEK_END );
 c_size = ftell( cf );
 rewind( cf );
 c_buf = malloc( c_size );
 fread( c_buf, 1, c_size, cf );
 fclose( cf );

	/* Remove spaces from config */

 for ( n = m = 0 ; n < c_size ; n++ ) {
	 if ( c_buf[ n ] != ' ' && c_buf[ n ] != '\t' && ! ( c_buf[ n ] == '\n' && c_buf[ m - 1 ] == '\n' )) c_buf[ m++ ] = tolower(c_buf[ n ]);
	}

	/* Parse entries from buffer */

 for ( n = 0 ; n < m ; n++ ) {
	 if ( ! memcmp( c_buf + n, "site{\n", 6 ) ) {
		 while ( n < m && c_buf[ n ] != '}' ) {
			 switch( c_buf[ n ] ) {
				 case ':':
					 v_start = n + 1;
					break;
				 case '\n':
					 if ( v_start > i_start ) {
						 t = c_buf + i_start;
						 l = v_start - i_start - 1 ;
						 if ( ! memcmp( t, "host", l ) ) snprintf( host, MAXHOSTNAMELEN, "%.*s", n - v_start, c_buf + v_start );
						 if ( ! memcmp( t, "port", l ) ) snprintf( port, 6, "%.*s", n - v_start, c_buf + v_start );
						}
					 i_start = n + 1;
					break;
				}
			 n++;
			}
		 if ( c_buf[ n ] == '}' ) {
			 site_ip[ target_hosts ] = malloc( strlen( host ) + 1 );
			 strcpy( site_ip[ target_hosts ], host );
			 site_port[ target_hosts ] = atoi( port );
			 target_hosts++;
			}
		}
	 if ( ! memcmp( c_buf + n, "settings{\n", 10 ) ) {
		 while ( n < m && c_buf[ n ] != '}' ) {
			 switch( c_buf[ n ] ) {
				 case ':':
					 v_start = n + 1;
					break;
				 case '\n':
					 if ( v_start > i_start ) {
						 t = c_buf + i_start;
						 l = v_start - i_start - 1 ;
						 if ( ! memcmp( t, "read-ident", l ) ) snprintf( readident, 5, "%.*s", n - v_start, c_buf + v_start );
						 if ( ! memcmp( t, "send-ident", l ) ) snprintf( sendident, 3, "%.*s", n - v_start, c_buf + v_start );
						 if ( ! memcmp( t, "bouncer-pass", l ) ) snprintf( pass, 24, "%.*s", n - v_start, c_buf + v_start );
						 if ( ! memcmp( t, "client-port", l ) ) snprintf( clientport, 6, "%.*s", n - v_start, c_buf + v_start );
						 if ( ! memcmp( t, "bouncer-port", l ) ) snprintf( bouncerport, 6, "%.*s", n - v_start, c_buf + v_start );
						 if ( ! memcmp( t, "distribution-mode", l ) ) snprintf( distrmode, 24, "%.*s", n - v_start, c_buf + v_start );
						 if ( ! memcmp( t, "shm-key", l ) ) snprintf( t_key, 12, "%.*s", n - v_start, c_buf + v_start );
						}
					 i_start = n + 1;
					break;
				}
			 n++;
			}
		 if ( c_buf[ n ] == '}' ) {
			 key = strtoul( t_key, (char **)NULL, 16 );
			 client_port = atoi( clientport );
			 bouncer_port = atoi( bouncerport );
			 bouncer_pass = strtoul( pass, (char **)NULL, 10 );
			 if ( ! strcmp( readident, "on" ) )
				 read_ident = 1;
				else if ( ! strcmp( readident, "idnt" ) ) 
				 read_ident = 2;
				else
				 read_ident = 0;
			 if ( ! strcmp( sendident, "on" ) )
				 send_ident = 1;
				else
				 send_ident = 0;
			 if ( ! strcmp( distrmode, "equal-transfers" ) )
				 distr_mode = 2;
				else if ( ! strcmp( distrmode, "equal-traffic" ) )
				 distr_mode = 1;
				else if ( ! strcmp( distrmode, "random" ) )
				 distr_mode = 4;
				else
				 distr_mode = 0;
			}
		}
	 while ( c_buf[ n ] != '\n' && n < m ) n++;
	}

 free( c_buf );

	/* Check if all variables were read from config */


 if ( bouncer_port > 0 &&
	client_port > 0 &&
	read_ident != -1 &&
	send_ident != -1 &&
	distr_mode != -1 &&
	key != 0 &&
	target_hosts > 0 ) return;

 printf( "dT-entry encountered fatal failure - config error!\n" );
 exit( 2 );
}







main ( int argc, char **argv ) {
 struct	sockaddr_in	client_entry_addr,
			bouncer_entry_addr,
			remote_client_addr,
			remote_server_addr;
 struct hostent		*hp;
 struct	cmd_t		command;
 struct rsp		response;
 fd_set			read_ds;
 int			server,
			client,
			bouncer_entry,
			client_entry,
			entry,
			sz,
			n,
			b_fd, g_fd;
 char			remote_client[ MAXHOSTNAMELEN ],
			ip[ 16 ];


	/* Startup parameters */

 if ( argc != 2 ) {
	 printf("dT-entry				(c) 2001 - dark0n3\n\n");
	 printf("Usage: %s <config>\n\n", argv[ 0 ] );
	 exit( 2 );
	}

 if ( argc == 2 ) {
	 read_config( argv[ 1 ] );
	}

	/* Set network parameters */


 sz = sizeof( remote_client_addr );

 bzero( &client_entry_addr, sz );
 bzero( &bouncer_entry_addr, sz );

 client_entry_addr.sin_addr.s_addr  =
 bouncer_entry_addr.sin_addr.s_addr = INADDR_ANY;
 client_entry_addr.sin_family  =
 bouncer_entry_addr.sin_family = AF_INET;
 client_entry_addr.sin_port  = htons( (unsigned)client_port );
 bouncer_entry_addr.sin_port = htons( (unsigned)bouncer_port );

 client_entry  = socket(PF_INET, SOCK_STREAM, 0 );
 bouncer_entry = socket(PF_INET, SOCK_STREAM, 0 );

	/* Bind to sockets */

 if ( bind( client_entry, (struct sockaddr *)&client_entry_addr, sz ) < 0 ) {
	 fprintf( stderr, "ERROR: Failed on client bind()!\n");
	 exit( 2 );
	}

 if ( bind( bouncer_entry, (struct sockaddr *)&bouncer_entry_addr, sz ) < 0 ) { 
	 fprintf( stderr, "ERROR: Failed on bouncer bind()!\n");
	 exit( 2 );
	}

	/* Create pipes and set them as nonblocking */


 pipe( pipes );

 g_fd = ( bouncer_entry > client_entry ? bouncer_entry : client_entry );
 g_fd = ( pipes[ 0 ] > g_fd ? pipes[ 0 ] : g_fd );

	/* Start listening */

 if ( listen( client_entry, 1024 ) != 0 ) {
	 fprintf( stderr, "ERROR: Failed on client listen()!\n" );
	 exit( 2 );
	}

 if ( listen( bouncer_entry, 1024 ) != 0 ) {
	 fprintf( stderr, "ERROR: Failed on bouncer listen()!\n" );
	 exit( 2 );
	}

 init_container();

 printf("dT-entry launched.\n");
 if ( fork() ) exit( 0 );

 log( "dTbentry session started\n" );

	/* Cycle until dead */

 while ( 1 ) {
	 FD_ZERO(&read_ds);
	 FD_SET(bouncer_entry, &read_ds);
	 FD_SET(client_entry, &read_ds);
	 FD_SET(pipes[0], &read_ds);

		/* Listen all connected traffic bouncers */

	 b_fd = g_fd;

	 for (n = 0; n < MAX_CONNECTIONS; n++ ) {
		if ( container[n].bnc_status >= ACTIVE ) {
			FD_SET(container[n].bnc_fd, &read_ds);
			b_fd = (container[n].bnc_fd > b_fd ? container[n].bnc_fd : b_fd);
			}
		}

	 select(b_fd + 1, &read_ds, 0, 0, 0);

		/* Listen for internal pipe */

	 if (FD_ISSET(pipes[0], &read_ds)) {
		read(pipes[0], &command, sizeof(struct cmd_t));
		direct_send_command( command.target, command.cmd, command.owner, command.hostname, command.port);
		}

		/* Accept new traffic bouncers, and request auth from them */
	
	 if ( FD_ISSET( bouncer_entry, &read_ds ) ) {
		entry = accept( bouncer_entry, (struct sockaddr *)&bouncer_entry_addr, &sz );
		strncpy( ip, inet_ntoa( bouncer_entry_addr.sin_addr ), sizeof( ip ) );
		log( "New traffic bouncer [%s] connected\n", ip );
		if ((n = add_container(entry, ip)) != -1) {
			direct_send_command(n, REQ_AUTH, NOONE, NULLSTR, 0);
			} else {
			close(entry);
			}
		}

		/* Read replies from bouncer, if not OFFLINE */

	 for ( entry = 0 ; entry < MAX_CONNECTIONS ; entry++ ) if ( container[ entry ].bnc_status != OFFLINE && FD_ISSET( container[ entry ].bnc_fd, &read_ds ) ) {
		n = read(container[entry].bnc_fd, &response, sizeof(struct rsp));
		if ( n <= 0 ) {
			container[entry].bnc_status = OFFLINE;
			container[entry].bnc_fd = -1;
			close(container[entry].bnc_fd);
			log("Traffic bouncer [%s] - connection lost\n", container[entry].bnc_ip);
			} else {
			switch(response.owner) {
				case STATS_BYTES:
					container[entry].bnc_bytes = response.value;
					break;
				case STATS_CONNECTIONS:
					container[entry].bnc_connections = response.value;
					break;
				case STATS_SPEED:
					container[entry].bnc_speed = response.value;
					break;
				case STATS_RESET:
					container[entry].bnc_reset = time(NULL) + response.value;
					break;
				case STATS_BYTES_LEFT:
					container[entry].bnc_bytes_left = response.value;
					if ( response.value == 0 )
						container[entry].bnc_status = CLOSED;
						else
						container[entry].bnc_status = ACTIVE;
					break;
				case PASSKEY:
					if ( bouncer_pass == response.value ) {
						container[entry].bnc_status = ACTIVE;
						direct_send_command(entry, REQ_STATS, NOONE, NULLSTR, 0);
						} else {
						log("Traffic bouncer [%s] authorisation failed!\n", container[entry].bnc_ip);
						close(container[entry].bnc_fd);
						container[entry].bnc_status = OFFLINE;
						}
					break;
				case PONG:
					log("PONG from traffic bouncer\n");
					break;
				default:
					log("Response for client's [%i] request arrived\n", response.owner);
					container[response.owner].bnc_reply = response.value;
					break;
				}
			}
		}

		/* Let new clients in, if there is still room */

	 if ( FD_ISSET( client_entry, &read_ds ) ) {

		 switch ( fork() ) {
			 case 0:
				for (n = 0; n < MAX_CONNECTIONS; n++ ) {
					if (container[n].user_online == 0) break;
					}

				if (container[n].user_online == 0) {
					container[n].user_online = 1;
					client_num = n;

					bzero(&remote_server_addr, sz);
					bzero(&remote_client_addr, sz);
					remote_server_addr.sin_family = AF_INET;

					client = accept(client_entry, (struct sockaddr *)&remote_client_addr, &sz);
					close(client_entry);

					remote_server_addr.sin_port = htons( (unsigned)site_port[ current_target ] );
					hp = gethostbyname( site_ip[ current_target ] );
					memcpy( &remote_server_addr.sin_addr, hp->h_addr, hp->h_length );

					hp = gethostbyaddr((char *)&remote_client_addr.sin_addr, sz, AF_INET);
					if ( hp )
						 strncpy( remote_client, hp->h_name, MAXHOSTNAMELEN );
					 	else
						 strncpy( remote_client, inet_ntoa( remote_client_addr.sin_addr ), MAXHOSTNAMELEN );

					server = socket( AF_INET, SOCK_STREAM, 0 );
					if ( connect( server, (struct sockaddr *) &remote_server_addr, sz ) == 0 ) {

						log( "New client [%s] connected with id %i\n", remote_client, client_num );

							/* Read & send Ident */

						ident_call(client, server, remote_client, &remote_client_addr, &client_entry_addr);

							/* Perform TLS login, if TLS enabled */

						if ( tls_login( client, server, remote_client ) != -1 ) {
							control_loop( client, server, client_num );
							}
						}
					container[ client_num ].user_online = 0;
					log( "Client [%s] logged off\n", remote_client );
					CLOSE( server );
					close( client );
					}
				 exit(0);
				break;
			 case -1:
				break;
			 default:
				 close( client );
				 if ( fork() ) exit ( 0 );
				break;
			}
		 current_target++;
		 if ( current_target >= target_hosts ) current_target = 0;
		}
	}
}
