#include <sys/stat.h>
#include <fcntl.h>
#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <netdb.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <resolv.h>
#include <net/if.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netinet/tcp.h>
#include "types.h"

int		port_max = 0;
int		port_min = 0;
int		random_ports = 0;
int		recv_buf = 0;
int		send_buf = 0;
int		next_reset = 0;
int		entries = 0;
int		last_port = 0;
int		reset_interval = 0;
int		retry_interval = 0;
int		excluded = 0;
int		entry_fd[MAX_ENTRIES];
int		pipes[2];
char		site_device[16];
char		client_device[16];
char		excluded_ip_list[MAX_EXCLUDED][16];
unsigned long	traffic_limit = 0;
		
struct	statistics	container;
struct	hostent		*hp;
struct	entry_list	entry[MAX_ENTRIES];
struct	tb_update	up;




/*
 * int bind_socket( int socket, char *host )
 *
 * Binds socket to chosen ip.
 *
 */
int
bind_socket(int socket, char *host) {
 struct	sockaddr_in	local_addr;

 bzero(&local_addr, sizeof(struct sockaddr_in));
 hp = gethostbyname(host);
 memcpy(&local_addr.sin_addr, hp->h_addr, hp->h_length);
 return bind(socket, (struct sockaddr *)&local_addr, sizeof(struct sockaddr_in));
}




/*
 * int open_new_port( int *skt )
 * 
 * Opens new port and starts listening on it. Returns port number
 * as integer and *skt is set to listening socket.
 *
 */
int
open_new_port(int *skt) {
 struct	timeval		tv;
 struct	sockaddr_in	server_addr;
 unsigned		port = port_min;
 int			server;

	/* Select port with chosen method */

 if (random_ports == ON) {
	gettimeofday(&tv,NULL);
	srandom(tv.tv_usec);
	} else {
	port = last_port;
	}

	/* Loop until bind to port succeeds */

 server = socket(PF_INET, SOCK_STREAM, 0);
 bzero(&server_addr, sizeof(server_addr));
 server_addr.sin_family = AF_INET;

 while ( 1 ) {
	if (random_ports == ON) {
		port = (unsigned)((port_max - port_min) * 1. * rand()/(RAND_MAX+1.0));
		port += port_min;
		} else {
		port++;
		if ( port > port_max ) port = port_min;
		}

	server_addr.sin_port = htons(port);

	if (bind(server, (struct sockaddr *)&server_addr, sizeof(server_addr)) == 0 ) { /* Make sure port is empty */
		break;
		}
	}
 listen(server, 128);
 last_port = port;
 *skt = server;
 return port;
}




/*
 * int is_excluded( char *ip )
 *
 * Checks wheter ip matches with excluded ip list or not.
 *
 */
int
is_excluded ( char *ip ) {
 int	n, m, l,
	any,
	matches;

 for ( l = 0 ; l < excluded ; l++ ) {
	 matches = 1;
	 any = 0;
	 for ( n = m = 0 ; n < 16 ; n++ ) {
		 if ( ip[n] == '.' ) {
			 if ( any ) m++;
			 any = 0;
			} else {
			 switch( excluded_ip_list[ l ][ m ] ) {
				case '*':
					 any = 2;
					break;
				case '?':
					 any = 1;
					break;
					}
			}
		 if ( ip[ n ] == excluded_ip_list[ l ][ m ] || any ) {
			 if ( any == 1 ) {
				 any = 0;
				}
			} else {
			 matches = 0;
			 break;
			}
		 if ( any == 0 ) m++;
		}
	 if ( matches ) return matches;
	}
 return 0;
}





/*
* void data_loop( int server_socket, int server )
*
* Transfers data until bitter end.
*
*/
void
data_loop(int server_socket, int server) {
 struct sockaddr_in	client_addr;
 struct timeval		start, stop;
 unsigned long		bytes = 0;
 fd_set			read_ds;
 int			client,
			b_fd,
			n,
			in,
			out,
			D_BUF_SIZE;
 char			*D_BUF;

 fcntl(server, 0);

	/* Wait up to 5 secs for client connect */

 stop.tv_sec = 5;
 stop.tv_usec = 0;

 FD_ZERO(&read_ds);
 FD_SET(server_socket, &read_ds);
 n = select(server_socket + 1, &read_ds, 0, 0, &stop);

 if (n > 0) {
	n = sizeof(client_addr);
	bzero(&client_addr, n);
	client = accept(server_socket, (struct sockaddr *)&client_addr, &n);
	close(server_socket);

	b_fd = client > server ? client : server;

		/* Allocate buffer */

        D_BUF_SIZE = recv_buf > send_buf ? recv_buf : send_buf;
	D_BUF = malloc(D_BUF_SIZE);

		/* Check bouncing direction */

	FD_ZERO(&read_ds);
	FD_SET(server, &read_ds);
	FD_SET(client, &read_ds);

	select(b_fd + 1, &read_ds, 0, 0, 0 );
	if (FD_ISSET(server, &read_ds)) {
		in = server;
		out = client;
		} else {
		in = client;
		out = server;
		}

		/* Update */

	up.command = SET_TRANSFERS;
	up.value = 1;
	write(pipes[1], &up, sizeof(struct tb_update));

	up.command = UPDATE_ALL;
	write(pipes[1], &up, sizeof(struct tb_update));

		/* Set send and recieve buffers */

	setsockopt(in, SOL_SOCKET, SO_RCVBUF, (int *)recv_buf, sizeof(int));
	setsockopt(out, SOL_SOCKET, SO_SNDBUF, (int *)send_buf, sizeof(int));

		/* Begin the transfer */

	gettimeofday(&start, (struct timezone *)0);
	while ((n = read(in, D_BUF, D_BUF_SIZE)) > 0 &&
		write(out, D_BUF, n) == n) {
		bytes += n;
		}
	gettimeofday(&stop, (struct timezone *)0 );

		/* Transfer end - update stats */

	up.command = SET_TRANSFERS;
	up.value = -1;
	write(pipes[1], &up, sizeof(struct tb_update));

	up.command = SET_TIME;
	up.value = (double)(((stop.tv_usec - start.tv_usec) / 1000000.) + (stop.tv_sec - start.tv_sec));
	write(pipes[1], &up, sizeof(struct tb_update));

	if (! is_excluded(inet_ntoa(client_addr.sin_addr))) {
		up.command = SET_LEFT;
		up.value = bytes / 1024;
		write(pipes[1], &up, sizeof(struct tb_update));
		}

	up.command = SET_TOTAL;
	up.value = bytes;
	write(pipes[1], &up, sizeof(struct tb_update));

	up.command = UPDATE_ALL;
	write( pipes[1], &up, sizeof(struct tb_update));

		/* Free buffer */

	free(D_BUF);

		/* Close sockets */

	close(client);
	} else {
	close(server_socket);
	}
 close(server);
}





/*
* void read_config( char *f )
*
* Reads dynamic variables from config file and 
* allocates memory for them
*
*/
void
read_config( char *config ) {
 FILE	*cf;
 int	c_size,
	n, m, l;
 char	*c_buf,
	*t,
 	host[ MAXHOSTNAMELEN ],
	port[ 6 ],
	pass[ 24 ],
	random[ 3 ],
	minport[ 6 ],
	maxport[ 6 ],
	retryinterval[ 6 ],
	resetinterval[ 6 ],
	trafficlimit[ 8 ],
	sitedev[ 16 ],
	clientdev[ 16 ],
	sendbuf[ 8 ],
	recvbuf[ 8 ];
 int	i_start = 0,
	v_start = 0;

 *site_device = *client_device = 0;

	/* Open Config file and buffer it */

 if ( ( cf = fopen( config, "r" ) ) == NULL ) {
	 printf( "dT-bnc encountered fatal failure - config not found!\n");
	 exit( 2 );   
	}
 fseek(cf, 0, SEEK_END);
 c_size = ftell(cf);
 rewind(cf);
 c_buf = malloc(c_size);
 fread(c_buf, 1, c_size, cf);
 fclose(cf);

	/* Remove spaces from config, and make it lowercase */

 for ( n = m = 0 ; n < c_size ; n++ ) {
	if ( c_buf[ n ] != ' ' && c_buf[ n ] != '\t' && ! ( c_buf[ n ] == '\n' && c_buf[ m - 1 ] == '\n' )) c_buf[ m++ ] = tolower(c_buf[ n ]);
	}
 for ( n = m + 1 ; n < c_size ; n++ ) c_buf[ n ] = 0;

	/* Parse entries from buffer */

 for ( n = 0 ; n < m ; n++ ) {
	if ( ! memcmp( c_buf + n, "entry{\n", 7 ) ) {
		while ( n < m && c_buf[ n ] != '}' ) {
			switch( c_buf[ n ] ) {
				case '=':
					v_start = n + 1;
					break;
				case '\n':
					if ( v_start > i_start ) {
						l = v_start - i_start - 1;
						if ( ! memcmp( c_buf + i_start, "host", l ) ) snprintf( host, MAXHOSTNAMELEN, "%.*s", n - v_start, c_buf + v_start );
						if ( ! memcmp( c_buf + i_start, "pass", l ) ) snprintf( pass, 24, "%.*s", n - v_start, c_buf + v_start );
						if ( ! memcmp( c_buf + i_start, "port", l ) ) snprintf( port, 6, "%.*s", n - v_start, c_buf + v_start );
						}
					i_start = n + 1;
					break;
				}
			n++;
			}
		if ( c_buf[n] == '}' ) {
			entry[ entries ].host = malloc( strlen( host ) + 1 );
			strcpy( entry[ entries ].host, host );
			entry[ entries ].port = atoi( port );
			entry[ entries ].pass = strtoul( pass, (char **)NULL, 10 );
			entries++;
			}
		}
	if ( ! memcmp(c_buf + n, "excluded{\n", 10 ) ) {
		while (n < m && c_buf[n] != '}') {
			switch(c_buf[n]) {
				case '=':
					v_start = n + 1;
					break;
				case '\n':
					if ( v_start > i_start ) {
						l = v_start - i_start - 1 ;
						if (! memcmp( c_buf + i_start, "ip", l)) snprintf(host, 16, "%.*s", n - v_start, c_buf + v_start);
						}
					i_start = n + 1;
					break;
				}
			n++;
			}
		if (c_buf[n] == '}') {
			strcpy(excluded_ip_list[excluded], host);
			excluded++;
			}
		}
	if ( ! memcmp(c_buf + n, "settings{\n", 10)) {
		while (n < m && c_buf[n] != '}') {
			switch(c_buf[n]) {
				case '=':
					v_start = n + 1;
					break;
				case '\n':
					if ( v_start > i_start ) {
						t = c_buf + i_start;
						l = v_start - i_start - 1 ;
						if (! memcmp(t, "random-ports", l ) ) snprintf(random, 3, "%.*s", n - v_start, c_buf + v_start);
						if (! memcmp(t, "max-port", l ) ) snprintf(minport, 6, "%.*s", n - v_start, c_buf + v_start);
						if (! memcmp(t, "min-port", l ) ) snprintf(maxport, 6, "%.*s", n - v_start, c_buf + v_start);
						if (! memcmp(t, "send-buf", l ) ) snprintf(sendbuf, 8, "%.*s", n - v_start, c_buf + v_start);
						if (! memcmp(t, "recv-buf", l ) ) snprintf(recvbuf, 8, "%.*s", n - v_start, c_buf + v_start);
						if (! memcmp(t, "site-device-ip", l ) ) snprintf(sitedev, 16, "%.*s", n - v_start, c_buf + v_start);
						if (! memcmp(t, "client-device-ip", l ) ) snprintf(clientdev, 16, "%.*s", n - v_start, c_buf + v_start);
						if (! memcmp(t, "retry-interval", l ) ) snprintf(retryinterval, 6, "%.*s", n - v_start, c_buf + v_start); 
						if (! memcmp(t, "traffic-limit", l ) ) snprintf(trafficlimit, 8, "%.*s", n - v_start, c_buf + v_start);
						if (! memcmp(t, "limit-reset", l ) ) snprintf(resetinterval, 6, "%.*s", n - v_start, c_buf + v_start);
						}
					i_start = n + 1;
					break;
				}
			n++;
			}
		if ( c_buf[n] == '}' ) {
			traffic_limit = strtoul(trafficlimit, (char **)NULL, 10 ) * 1024;
			retry_interval = atoi(retryinterval);
			reset_interval = atoi(resetinterval);
			port_min = atoi(minport);
			port_max = atoi(maxport);
			recv_buf = atoi(recvbuf);
			send_buf = atoi(sendbuf);
			strcpy(client_device, clientdev);
			strcpy(site_device, sitedev);
			if (! strcasecmp(random, "on")) {
				random_ports = 1;
				} else {
				random_ports = 0;
				}
			}
		}
	while (c_buf[n] != '\n' && n < m) n++;
	}

 free(c_buf);

	/* Check if all variables were read from config */

 last_port = port_min;

 next_reset = time( NULL ) + reset_interval * 60;

 if ( send_buf > 0 &&
	recv_buf > 0 &&
	port_min > 0 &&
	port_max > 0 &&
	random_ports != -1 &&
	retry_interval > 0 &&
	entries > 0 &&
	*client_device &&
	*site_device ) return;

 printf("dT-bnc encountered fatal failure - config error!\n");
 exit(2);
}




/*
* void send_response( int fd, int owner, unsigned long value )
*
* Sends response to entry.
*
*/
void
send_response(int bnc, int owner, unsigned long value) {
 struct rsp	response;

	/* Send response to entry */

 response.owner = owner;
 response.value = value;
 write(entry_fd[bnc], &response, sizeof(struct rsp));
}







main (int argc, char **argv) {
 struct	sockaddr_in	entry_addr;
 struct sockaddr_in	remote_addr;
 struct	cmd		command;
 struct	conn		connection[MAX_CONNECTIONS];
 struct timeval		retry;
 fd_set			read_ds;
 int			remote;
 int			len;
 int			b_fd;
 int			n, m;
 int			fd;
 int			port = 0;
 int			user = 0;

	/* Read configuration file */

 if (argc != 2) {
	exit(2);
	}

 read_config(argv[1]);

	/* Initialize arrays */

 container.active_transfers = 0;
 container.total_bytes_transfered = 0;
 container.transfer_time = 0;
 container.bytes_left = traffic_limit;

 pipe(pipes);

 bzero(&entry_fd, sizeof(int) * entries);
 bzero(&retry, sizeof(struct timeval));

 for ( user = 0 ; user < MAX_CONNECTIONS ; user++ ) {
	connection[user].owner = -1;
	connection[user].listen_socket = -1;
	connection[user].source_socket = -1;
	}

 printf("dT-bnc launched.\n" );

 if (fork() != 0) exit(0);

 while ( 1 ) {
	 FD_ZERO(&read_ds);
	 FD_SET(pipes[0], &read_ds);

	 b_fd = pipes[0];

			/* Try to connect to entries if retry counter is zero or less - otherwise just listen */

	 for ( n = 0 ; n < entries ; n++ ) {
		fd = entry_fd[n];

		if ( fd > 0 ) {
			FD_SET (fd, &read_ds);
			b_fd = (fd > b_fd ? fd : b_fd);
			} else if (retry.tv_sec <= 0 && retry.tv_usec <= 0) {
			fd = socket(AF_INET, SOCK_STREAM, 0); 
			bind_socket(fd, client_device);

			bzero(&entry_addr, sizeof(struct sockaddr_in));
			entry_addr.sin_family = AF_INET;
			entry_addr.sin_port = htons((unsigned)entry[n].port);
			hp = gethostbyname(entry[n].host);
			memcpy(&entry_addr.sin_addr, hp->h_addr, hp->h_length);

			fcntl(fd, O_NONBLOCK);
			connect(fd, (struct sockaddr *)&entry_addr, sizeof(entry_addr));
			fcntl(fd, 0);

			FD_SET(fd, &read_ds);
			b_fd = (fd > b_fd ? fd : b_fd);
			entry_fd[n] = fd;
			}
		}

			 /* Reset retry counter */

	 if ( retry.tv_sec <= 0 && retry.tv_usec <= 0 ) {
		retry.tv_sec = retry_interval;
		retry.tv_usec = 0;
		}

			/* Wait until data arrives - or retry counter goes to zero */

	 select(b_fd + 1, &read_ds, 0, 0, &retry);

			/* Read data coming from entry */

	 for (n = 0; n < entries; n++) if (entry_fd[n] > 0 && FD_ISSET(entry_fd[n], &read_ds)) {
		len = read(entry_fd[n], &command, sizeof(struct cmd));
			/* Connection to entry lost or failed */

		if (len <= 0) {
			close(entry_fd[n]);
			entry_fd[n] = -1;
			continue;
			}

			/* Search user from command buffer, if not there - create new entry or skip if buffer full */

		if ( command.owner != NOONE ) {
			for (user = 0; user < MAX_CONNECTIONS; user++) if (connection[user].owner == command.owner && connection[user].entry == n) break;
			if ( user == MAX_CONNECTIONS ) {
				for (user = 0; user < MAX_CONNECTIONS; user++) if ( connection[user].owner == -1 ) break;
				if (user == MAX_CONNECTIONS) {
					continue; /* This should not happend */
					}
				connection[user].owner = command.owner;
				connection[user].entry = n;
				}
			}

		 switch (command.cmd) {

				/* Entry sent connect (for PASV/PORT) request */

			 case REQ_PORT:
				port = 1;

			 case REQ_PASV:
				send_response(n, command.owner, open_new_port(&connection[user].listen_socket));

				bzero(&remote_addr, sizeof(struct sockaddr_in));
				remote_addr.sin_family = AF_INET;
				remote_addr.sin_port = htons((unsigned)command.port);
				hp = gethostbyname(command.hostname);
				memcpy(&remote_addr.sin_addr, hp->h_addr, hp->h_length);

				remote = socket(AF_INET, SOCK_STREAM, 0);
				fcntl(remote, O_NONBLOCK);

				if ( port == 0 ) {
					bind_socket(remote, site_device);
					} else {
					bind_socket(remote, client_device);
					}

				connect(remote, (struct sockaddr *)&remote_addr, sizeof(remote_addr));
				connection[user].source_socket = remote;
				port = 0;
				break;

				/* Entry wants to begin the transfer */

			 case REQ_ACCEPT:
				 switch (fork()) {
					case 0:
						data_loop(connection[user].listen_socket, connection[user].source_socket);
						exit(0);
						break;
					case -1:
						break;
					default:
						if (fork()) exit(0);
						break;
					}
				/* We can safely clear info for forked process */

				/* Entry has sent clear request */

			 case REQ_CLEAR:
				close(connection[user].listen_socket);
				close(connection[user].source_socket);
				connection[user].listen_socket = -1;
				connection[user].source_socket = -1;
				connection[user].entry = -1;
				connection[user].owner = -1;
				break;

				/* Entry requested for our stats */

			 case REQ_STATS:
				send_response(n, STATS_BYTES, container.total_bytes_transfered / 1024);
				send_response(n, STATS_BYTES_LEFT, container.bytes_left);
				send_response(n, STATS_CONNECTIONS, container.active_transfers);
				send_response(n, STATS_SPEED, container.total_bytes_transfered / container.transfer_time);
				send_response(n, STATS_RESET, next_reset - time(NULL));
				break;

				/* Entry requested authentication */

			 case REQ_AUTH:
				send_response(n, PASSKEY, entry[n].pass);
				break;
			}
		}

			/* Time to reset stats */

	 if ( time(NULL) >= next_reset ) {
		container.bytes_left = traffic_limit;
		next_reset = time(NULL) + 60 * reset_interval;
		for (n = 0; n < entries; n++ ) if (entry_fd[n] > 0) {
				send_response(n, STATS_BYTES_LEFT, container.bytes_left);
				send_response(n, STATS_RESET, next_reset - time(NULL));
				}
		} 


			/* Read data coming from internal pipes */

	 if (FD_ISSET(pipes[0], &read_ds)) {
		 read(pipes[0], &up, sizeof(struct tb_update));
		 switch(up.command) {
			case SET_TRANSFERS:
				container.active_transfers += up.value;
				break;
			case SET_TOTAL:
				container.total_bytes_transfered += up.value;
				break;
			case SET_TIME:
				container.transfer_time += up.value;
				break;
			case SET_LEFT:
				container.bytes_left -= container.bytes_left > up.value ? up.value : container.bytes_left;
				break;
			case UPDATE_ALL:
				for (n = 0 ; n < entries ; n++) {
					send_response(n, STATS_BYTES, container.total_bytes_transfered / 1024);
					send_response(n, STATS_BYTES_LEFT, container.bytes_left);
					send_response(n, STATS_CONNECTIONS, container.active_transfers);
					send_response(n, STATS_SPEED, container.total_bytes_transfered / container.transfer_time);
					}
				break;
			}
		}
	}
}

