#define PACKET_SIZE	35
#define BUF_SIZE	2048	/* Buffer might grow up to BUF_MAX_SIZE, this is the initial value */
#define BUF_MAX_SIZE	65535	/* Kill user if this exceeds, buffer size is only grown if response added by bnc does not fit in current buff */

#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/time.h>
#include <netdb.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <resolv.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <resolv.h>
#include <stdarg.h>
#include <signal.h>

#ifdef TLS
#include <openssl/ssl.h>
#include <openssl/x509v3.h>
#include <openssl/err.h>
#include <openssl/bio.h>
#include <openssl/rand.h>
#else
#include <errno.h>
#endif


#include "global.h"
#include "config-reader.h"

struct _TB {
	struct sockaddr_in	addr;
	int			sock;
	unsigned char		status;
	unsigned long		kbytes_total;
	unsigned long		kbytes_left;
	unsigned long		kbytes_xfer;
	unsigned long		time_total;
	time_t			next_reset;
	time_t			last_use;
	unsigned short		transfers_active;
	unsigned char		packet[PACKET_SIZE];
	unsigned char		packet_len;
	struct _TB		*next;
	struct _TB		*prev;
};


struct _CONFIG {
	struct sockaddr_in	site_addr;
	char			*key;
	unsigned char		key_len;
	unsigned		connect_timeout;
	unsigned		ident_timeout;
	unsigned char		read_ident;
	unsigned short		user_port;
	unsigned short		bnc_port;
	unsigned char		distr_mode;
	struct _TB		*last_bnc;
};


struct _USER {
	struct sockaddr_in	addr;
	int			ident_sock;
	int			user_sock;
	int			site_sock;
	unsigned short		uid;
	unsigned char		status;
	struct _TB		*bnc;

	unsigned char		data_status;
	time_t			timeout;
	char			ident[256];

#ifdef TLS
	unsigned char		ssl_status;
	SSL			*ssl;
#endif

	int			s_buf_size;  /* Size of s_buf */
	int			u_buf_size;  /* Size of u_buf */		
	unsigned char		*u_buf;
	unsigned char		*s_buf;

	char			read_u_ignore;
	char			read_s_ignore;
	unsigned short		write_s_len; /* Bytes to write to site socket from u_buf */
	unsigned short		write_u_len; /* Bytes to write to user socket from s_buf */
	unsigned short		read_u_len;  /* Bytes read from user socket to u_buf */
	unsigned short		read_s_len;  /* Bytes read from site socket to s_buf */

	struct _USER		*next;
	struct _USER		*prev;
};

#ifdef TLS
#define READ(s,f,b,c)	tls_read(s,f,b,c)
#define WRITE(s,f,b,c)	tls_write(s,f,b,c)

extern DH			*dh;
extern SSL_CTX			*ssl_ctx;
extern char			*tls_rsa_key_file;
extern char			*tls_rsa_cert_file;
extern char			*tls_dsa_key_file;
extern char			*tls_dsa_cert_file;
extern char			*tls_dhparam_file;
extern char			*tls_rand_file;

#else
#define READ(s,f,b,c)	read(f,b,c)
#define WRITE(s,f,b,c)	write(f,b,c)
#endif

#ifdef BLOWFISH
char		config_key[128];
char		config_key_len = 0;
#endif

struct _TB	*tb;
struct _TB	*last_bnc;
struct _USER	*user;
struct _USER	*last_user;
struct _CONFIG	config;
unsigned char	packet[PACKET_SIZE];
extern	int	errno;
time_t		now;
time_t		next_event;
int		sz = sizeof(struct sockaddr_in);
unsigned char	magic_key_len;
char		*magic_key;
static char	copyright[] = "Copyright (c) dark0n3 2001 - 2002 <darkone@glftpd.com>";
static char	stat_header[] =
	"200-                                dTbnc statistics                dark0n3 - 2002\r\n"
	"200- ._______________________________________________________________________________.\r\n"
	"200- |       IP        | STATUS |  LEFT   |   TOTAL  |    AVG      | USERS |  RESET  |\r\n"
	"200- |-----------------|--------|---------|----------|-------------|-------|---------|\r\n"
	"200- |                 |        |         |          |             |       |         |\r\n";
static char	stat_footer[] =
	"200- |_________________|________|_________|__________|_____________|_______|_________|\r\n"
	"200-\r\n"
	"200 Command Succesful.\r\n";



#define add_timer(t)	if (t < next_event || next_event == -1) next_event = t;



int add_to_s_buf(struct _USER *usr, char *s, int s_size) {
 if ( s_size + usr->read_s_len > usr->s_buf_size ) {
	if ( usr->s_buf_size += s_size + 512 > BUF_MAX_SIZE ) {
		return -1;
		}
	usr->s_buf = realloc(usr->s_buf, usr->s_buf_size);
	}
 memcpy(usr->s_buf + usr->read_s_len, s, s_size);
 usr->write_u_len += s_size;
 usr->read_s_len += s_size;
 return 0;
}




int add_to_u_buf(struct _USER *usr, char *s, int s_size) {
 if ( s_size + usr->read_u_len > usr->u_buf_size ) {
	if ( usr->u_buf_size += s_size + 512 > BUF_MAX_SIZE ) {
		return -1;
		}
	usr->u_buf = realloc(usr->u_buf, usr->u_buf_size);
	}
 memcpy(usr->u_buf + usr->read_u_len, s, s_size);
 usr->write_s_len += s_size;
 usr->read_u_len += s_size;
 return 0;
}





struct _TB *get_next_bnc() {
 unsigned short		n;
 struct _TB		*bnc;
 struct _TB		*chosen;
 
 switch ( config.distr_mode ) {

	/****** Select bouncer using round-robin method ******/

	case SELECT_ROBIN:
		bnc = (config.last_bnc != NULL ? config.last_bnc : tb);
		while ((bnc = (bnc->next != NULL ? bnc->next : tb->next)) != NULL) {
			if ( bnc->status == B_ACTIVE || bnc == config.last_bnc ) {
				break;
				}
			}
		config.last_bnc = chosen = bnc;
		break;

	/****** Select bouncer that has done least traffic ******/

	case SELECT_TRAFFIC:
		for ( chosen = tb->next ; chosen->next != NULL && chosen->status != B_ACTIVE ; chosen = chosen->next );
		for ( bnc = chosen->next ; bnc != NULL ; bnc = bnc->next ) {
			if ( bnc->status == B_ACTIVE ) {
				if ( chosen->kbytes_xfer > bnc->kbytes_xfer ) {
					chosen = bnc;
					} else if ( chosen->kbytes_xfer == bnc->kbytes_xfer && bnc->last_use <= chosen->last_use ) {
					chosen = bnc;
					}
				}
			}
		break;

	/****** Select bouncer that has least active connections ******/

	case SELECT_CONNECTIONS:
		for ( chosen = tb->next ; chosen->next != NULL && chosen->status != B_ACTIVE ; chosen = chosen->next );
		for ( chosen = chosen->next ; bnc != NULL ; bnc = bnc->next ) {
			if ( bnc->status == B_ACTIVE ) {
				if ( chosen->transfers_active > bnc->transfers_active ) {
					chosen = bnc;
					} else if ( chosen->transfers_active == bnc->transfers_active && bnc->last_use <= chosen->last_use ) {
					chosen = bnc;
					}
				}
			}
		break;

	/****** Choose bouncer randomly - eh eh ******/

	case SELECT_RANDOM:
		n = 0;
		for ( bnc = tb->next ; bnc != NULL ; bnc = bnc->next ) {
			if ( bnc->status == B_ACTIVE ) {
				n++;
				}
			}

		if ( n > 0 ) {
			n = (unsigned short)(n * 1. * rand() / (RAND_MAX + 1.0));

			for ( bnc = tb->next ; bnc != NULL ; bnc = bnc->next ) {
				if ( bnc->status == B_ACTIVE ) {
					if ( n == 0 ) {
						break;
						}
					n--;
					}
				}
			chosen = bnc;
			} else {
			chosen = NULL;
			}
		break;
	}


	/****** Return NULL if no active bouncer got selected ******/

 if ( chosen != NULL ) {
	if ( chosen->status == B_ACTIVE ) {
		chosen->last_use = now;
		} else {
		chosen = NULL;
		}
	}


 return chosen;
}







char my_time[60];
char *neat_time(long sec) {
 unsigned char		hour = 0;
 unsigned char		min = 0;

 while ( sec >= 3600 ) {
	sec -= 3600;
	hour++;
	}

 while ( sec >= 60 ) {
	sec -= 60;
	min++;
	}

 if ( hour ) {
	sprintf(my_time, "%uh %um", hour, min);
	} else {
	sprintf(my_time, "%um %us", min, sec);
	}

 return my_time;
}




int write_tb_list(struct _USER *usr) {
 struct	_TB	*bnc;
 char		row[100];
 char		ip[16];
 unsigned char	t_ip[4];
 unsigned char	r_len;
 unsigned short	n;

 add_to_s_buf(usr, stat_header, sizeof(stat_header) - 1);

 for ( bnc = tb->next ; bnc != NULL ; bnc = bnc->next ) {
	memcpy(t_ip, &bnc->addr.sin_addr, 4);
	sprintf(ip, "%u.%u.%u.%u", *t_ip, *(t_ip + 1), *(t_ip + 2), *(t_ip + 3));
	switch ( bnc->status ) {
		case B_WAIT_AUTH:
			r_len = sprintf(row, "200- | %-15s |  AUTH  |   0.00G |     0.0G |     0.0KB/s |     0 |      0s |\r\n", ip);
			break;
		case B_ACTIVE:
			r_len = sprintf(row, "200- | %-15s | ACTIVE | %6.2fG | %7.1fG | %7.1fKB/s | %5u | %7s |\r\n", ip, (bnc->kbytes_left >> 10) / 1024., (bnc->kbytes_total >> 10) / 1024., bnc->kbytes_total * 1. / bnc->time_total, bnc->transfers_active, neat_time(bnc->next_reset - now));
			break;
		case B_CLOSED:
			r_len = sprintf(row, "200- | %-15s | CLOSED |   0.00G | %7.1fG | %7.1fKB/s |     0 | %7s |\r\n", ip, (bnc->kbytes_total >> 10) / 1024., bnc->kbytes_total * 1. / bnc->time_total, neat_time(bnc->next_reset - now));
			break;
		}
	if ( add_to_s_buf(usr, row, r_len) == -1 ) {
		return -1;
		}
	}

 add_to_s_buf(usr, stat_footer, sizeof(stat_footer) - 1);

 return 0;
}






/*
 *
 * Kills user
 *
 */
struct _USER *kill_user(struct _USER *usr) {
 struct _USER	*prev_user;

 if ( usr->bnc > 0 ) {
	*(packet + 2) = REQ_CLEAN;
	memcpy(packet + 3, &usr->uid, 2);
	write(usr->bnc->sock, packet, PACKET_SIZE);
	}

	/****** Close site socket ******/

 if ( usr->site_sock > 0 ) {
	close(usr->site_sock);
	}

	/****** Close user socket ******/

 if ( usr->user_sock > 0 ) {
	close(usr->user_sock);
	}

	/****** TLS mem free ******/

#ifdef TLS
 if ( usr->ssl != NULL ) {
	SSL_free(usr->ssl);
	}
#endif

 prev_user = usr->prev;
 if ( usr->next ) {
	prev_user->next = usr->next;
	(usr->next)->prev = prev_user;
	} else {  
	prev_user->next = NULL;
	last_user = prev_user;
	}

 free(usr);

 return prev_user;
}







/*
 *
 * Kills traffic bouncer
 *
 */
struct _TB *kill_bouncer(struct _TB *bnc) {
 struct _TB	*prev_bnc;
 struct _USER	*usr;
 int		s = bnc->sock;
 int		n;

	/****** Close bouncer socket ******/

 if ( s > 0 ) {
	close(s);
	for ( usr = user->next ; usr != NULL ; usr = usr->next ) {
		if ( usr->bnc == bnc ) {
			usr->data_status = D_INACTIVE;
			usr->bnc = NULL;
			}
		}
	}

	/****** Free data ******/

 prev_bnc = bnc->prev;

 if ( config.last_bnc == bnc ) {
	config.last_bnc = prev_bnc;
	}

 if ( bnc->next ) {
	prev_bnc->next = bnc->next;
	(bnc->next)->prev = prev_bnc;
	} else {
	prev_bnc->next = NULL;
	last_bnc = prev_bnc;
	}

 free(bnc);

 return prev_bnc;
}








int get_config() {
 struct hostent		*hp;
 char			*key;
 char			*value;
 int			l;

 while ((key = seek_config_next_branch()) != NULL) {
	if ( ! memcmp(key, "global", 7)) {

		/****** Get user listening port ******/

		if ((value = seek_config_key("user port")) == NULL ||
		    (config.user_port = atoi(value)) == 0) {
			printf("FATAL: 'user port' was not properly defined in global!\n");
			return -1;
			}

		/****** Get bouncer listening port ******/

		if ((value = seek_config_key("bouncer port")) == NULL ||
		    (config.bnc_port = atoi(value)) == 0) {
			printf("FATAL: 'bouncer port' was not properly defined in global!\n");
			return -1;
			}

		/****** Get traffic distribution mode ******/

		if ((value = seek_config_key("distribution mode")) == NULL ) {
			printf("FATAL: 'distribution mode' was not properly defined in global!\n");
			return -1;
			}
		if ( ! strcasecmp(value, "round-robin")) {
			config.distr_mode = SELECT_ROBIN;
			} else if ( ! strcasecmp(value, "equal traffic")) {
			config.distr_mode = SELECT_TRAFFIC;
			} else if ( ! strcasecmp(value, "equal connections")) {
			config.distr_mode = SELECT_CONNECTIONS;
			} else if ( ! strcasecmp(value, "random")) {
			config.distr_mode = SELECT_RANDOM;
			} else {
			printf("FATAL: 'distribution mode' was not properly defined in global!\n");
			return -1;
			}

		/****** Get site address ******/

		if ((value = seek_config_key("site host")) == NULL ||
		    (hp = gethostbyname(value)) == NULL) {
			printf("FATAL: 'site host' was not properly defined in global!\n");
			return -1;
			}
		memcpy(&config.site_addr.sin_addr, hp->h_addr, hp->h_length);


		/****** Get site port ******/

		if ((value = seek_config_key("site port")) == NULL ||
		    (config.site_addr.sin_port = htons(atoi(value))) == 0) {
			printf("FATAL: 'site port' was not properly defined in global!\n");
			return -1;
			}
		config.site_addr.sin_family = AF_INET;

		/****** Get bouncer pass ******/

		if ((value = seek_config_key("bouncer pass")) == NULL ||
		    (config.key_len = strlen(value)) == 0 ) {
			printf("FATAL: 'bouncer pass' was not properly defined in global!\n");
			return -1;
			}
		config.key_len = (config.key_len < PACKET_SIZE - 4 ? config.key_len : PACKET_SIZE - 4);
		config.key = malloc(config.key_len);
		memcpy(config.key, value, config.key_len);

		magic_key = malloc(config.key_len + 9);
		magic_key_len = sprintf(magic_key, "dtbnc %.*s", config.key_len, config.key);

		
		/****** Get connect timeout ******/

		if ((value = seek_config_key("connect timeout")) == NULL ||
		    (config.connect_timeout = atoi(value)) == 0 ) {
			printf("FATAL: 'connect timeout' was not properly defined in global!\n");
			return -1;
			}

		/****** Get ident timeout ******/

		if ((value = seek_config_key("ident timeout")) == NULL ||
		    (config.ident_timeout = atoi(value)) == 0 ) {
			printf("FATAL: 'ident timeout' was not properly defined in global!\n");
			return -1;
			}

		/****** Get ident mode ******/

		if ((value = seek_config_key("read ident")) == NULL ) {
			printf("FATAL: 'read ident' was not properly defined in global!\n");
			return -1;
			}
		if (! strcasecmp(value, "true")) {
			config.read_ident = 1;
			} else if (! strcasecmp(value, "false")) {
			config.read_ident = 0;
			} else {
			printf("FATAL: 'read ident' was not properly defined in global!\n");
			return -1;
			}

#ifdef TLS
		/****** Get TLS stuff ******/

		if (value = seek_config_key("dsa cert")) {
			l = strlen(value) + 1;
			tls_dsa_cert_file = malloc(l);
			memcpy(tls_dsa_cert_file, value, l);
			}
		if (value = seek_config_key("rsa cert")) {
			l = strlen(value) + 1;
			tls_rsa_cert_file = malloc(l);
			memcpy(tls_rsa_cert_file, value, l);
			}
		if (value = seek_config_key("dsa key")) {
			l = strlen(value) + 1;
			tls_dsa_key_file = malloc(l);
			memcpy(tls_dsa_key_file, value, l);
			}
		if (value = seek_config_key("rsa key")) {
			l = strlen(value) + 1;
			tls_rsa_key_file = malloc(l);
			memcpy(tls_rsa_key_file, value, l);
			}
		if (value = seek_config_key("dh param")) {
			l = strlen(value) + 1;
			tls_dhparam_file = malloc(l);
			memcpy(tls_dhparam_file, value, l);
			}
		if (! tls_dhparam_file || ! ((tls_rsa_cert_file && tls_rsa_key_file) || (tls_dsa_cert_file && tls_dsa_key_file))) {
			printf("FATAL: 'tls options' were not properly defined in global!\n");
			return -1;
			}
#endif

		return 0;
		} else {
		printf("WARNING: Obsolete entry in config: %s\n", key);
		}
	}

 printf("FATAL: 'global' branch not found in config!\n");
 return 1;
}









/*
 * 
 * Parse user buffer for new commands
 *
 */
int manage_new_u_data(struct _USER *usr, ssize_t count) {
 struct _TB	*bnc;
 unsigned char	port_value[6];
 unsigned short	newlines = 0;
 unsigned short	check_len, real_len;
 int		event;
 char		*p;
 char		killed;
 char		*line_p;
 char		*end_p;
 

 usr->read_u_len += count;

 p = usr->u_buf + usr->write_s_len;
 end_p = usr->u_buf + usr->read_u_len;

 for ( line_p = p ; p < end_p ; p++ ) {
	if ( *p == '\n' ) {
		event = EVENT_OK;
		real_len = p - line_p + 1;
		if ( real_len > 1 && *(p - 1) == '\r' ) {
			check_len = real_len - 2;
			} else {
			check_len = real_len - 1;
			}
		if ( usr->read_u_ignore == 1 ) {
			usr->read_u_ignore = 0;
			event = EVENT_HIDE;
			} else if ( check_len == 4 ) {
			if (! strncasecmp(line_p, "LIST", 4) ||
			    ! strncasecmp(line_p, "NLST", 4)) {
				if ( usr->data_status == D_ACTIVE ) {
					event = EVENT_OK;
					} else {
					event = EVENT_FAIL_NOBNC;
					}
				}
			} else if (check_len > 5) {
			if ( ! strncasecmp(line_p, "PORT ", 5) && sscanf(line_p + 5, "%hu,%hu,%hu,%hu,%hu,%hu", port_value, port_value + 1, port_value + 2, port_value + 3, port_value + 4, port_value + 5) == 6 ) {
				if ((bnc = get_next_bnc()) != NULL ) {
					killed = 0;

					if ( usr->bnc != NULL ) {
						*(packet + 2) = REQ_CLEAN;
						memcpy(packet + 3, &usr->uid, 2);
						if ( write(usr->bnc->sock, packet, PACKET_SIZE) != PACKET_SIZE ) {
							if ( usr->bnc == bnc ) {
								killed = 1;
								}
							kill_bouncer(usr->bnc);
							}
						}

					if ( killed == 0 ) {
						*(packet + 2) = REQ_PORT;
						memcpy(packet + 3, port_value, 6);
						memcpy(packet + 9, &usr->uid, 2);
						if ( write(bnc->sock, packet, PACKET_SIZE) != PACKET_SIZE ) {
							kill_bouncer(usr->bnc);
							event = EVENT_FAIL_NOBNC;
							} else {
							usr->bnc = bnc;
							event = EVENT_HIDE;
							}
						} else {
						event = EVENT_FAIL_NOBNC;
						}
					} else {
					event = EVENT_FAIL_NOBNC;
					}
				} else if (*(line_p + 4) == ' ') {
				if (! strncasecmp(line_p, "STOR", 4) ||
				    ! strncasecmp(line_p, "APPE", 4) ||
				    ! strncasecmp(line_p, "STOU", 4)) {
					if ( usr->data_status == D_ACTIVE ) {
						event = EVENT_OK;
						} else {
						event = EVENT_FAIL_NOBNC;
						}
					} else if (! strncasecmp(line_p, "RETR", 4)) {
					if ( usr->data_status == D_ACTIVE ) {
						event = EVENT_OK;
						} else {
						event = EVENT_FAIL_NOBNC;
						}
					} else if (! strncasecmp(line_p, "LIST", 4) ||
						   ! strncasecmp(line_p, "NLST", 4)) {
					if ( usr->data_status == D_ACTIVE ) {
						event = EVENT_OK;
						} else {
						event = EVENT_FAIL_NOBNC;
						}
					} else if (check_len == 8 && ! strncasecmp(line_p, "AUTH TLS", 8)) {
#ifdef TLS
					if ( usr->ssl == NULL ) {
						usr->ssl = SSL_new(ssl_ctx);
						SSL_set_fd(usr->ssl, usr->user_sock);
						if ( add_to_s_buf(usr, "234 AUTH TLS sucessful.\r\n", 25) == -1 ) {
							return -1;
							}
						usr->status = U_WAIT_SSL_ACCEPT;
						usr->ssl_status = S_WANT_WRITE;
						} else {
						if ( add_to_s_buf(usr, "500 TLS already enabled.\r\n", 26) == -1 ) {
							return -1;
							}
						}
#else
					if ( add_to_s_buf(usr, "500 AUTH TLS not implemented.\r\n", 31) == -1 ) {
							return -1;
							}
#endif
					event = EVENT_HIDE;
					} else if (! strncasecmp(line_p, "PBSZ", 4)) {
					if ( add_to_s_buf(usr, "500 PBSZ not supported.\r\n", 25) == -1 ) {
						return -1;
						}
					event = EVENT_HIDE;
					} 
				} else if (check_len == magic_key_len && ! memcmp(line_p, magic_key, check_len)) {
				if ( write_tb_list(usr) == -1 ) {
					return -1;
					}
				event = EVENT_HIDE;
				}
			}

		switch ( event ) {
			case EVENT_FAIL_NOBNC:
				if ( add_to_s_buf(usr, "520 No active trafficbouncers.\r\n", 32) == -1 ) {
					return -1;
					}
			case EVENT_HIDE:
				if ( p + 1 < end_p ) {
					memmove(line_p, p + 1, end_p - p);
					}
				usr->read_u_len -= real_len;
				break;
			case EVENT_OK:
				usr->write_s_len += real_len;
				break;
			}

		line_p = p + 1;
		newlines++;
		}
	}


	/****** Check if user sent too long command ******/

 if ( newlines == 0 ) {
	if ( usr->read_u_ignore == 1 ) {
		usr->read_u_len = 0;
		} else if ( usr->write_s_len == 0 && usr->read_u_len == BUF_SIZE ) {
		usr->read_u_len = 0;
		if ( add_to_s_buf(usr, "500 Command too long.\r\n", 23) == -1 ) {
			return -1;
			}
		usr->read_u_ignore = 1;
		}
	}
 return 0;
}










/*
 *
 * Parse site buffer for new responses
 *
 */
int manage_new_s_data(struct _USER *usr, ssize_t count) {
 struct _TB	*bnc;
 unsigned char	pasv_value[6];
 unsigned short	newlines = 0;
 int		event;
 unsigned short	check_len, real_len;
 char		*p;
 char		killed;
 char		*line_p;
 char		*end_p, *real_end_p;
 

 usr->read_s_len += count;

 p = usr->s_buf + usr->write_u_len;
 real_end_p = end_p = usr->s_buf + usr->read_s_len;

 for ( line_p = p ; p < end_p ; p++ ) {
	if ( *p == '\n' ) {
		event = EVENT_OK;
		real_len = p - line_p + 1;
		if ( real_len > 1 && *(p - 1) == '\r' ) {
			check_len = real_len - 2;
			} else {
			check_len = real_len - 1;
			}

		if ( usr->read_s_ignore == 1 ) {
			usr->read_s_ignore = 0;
			} else if (check_len > 5) {
			if ( ! memcmp(line_p, "150 ", 4) ) {
				usr->data_status = D_INACTIVE;
				} else if ( sscanf(line_p, "227 Entering Passive Mode (%hu,%hu,%hu,%hu,%hu,%hu)", pasv_value, pasv_value + 1, pasv_value + 2, pasv_value + 3, pasv_value + 4, pasv_value + 5) == 6 ) {
				if ((bnc = get_next_bnc()) != NULL ) {
					killed = 0;
					
					if ( usr->bnc != NULL ) {
						*(packet + 2) = REQ_CLEAN;
						memcpy(packet + 3, &usr->uid, 2);
						if ( write((usr->bnc)->sock, packet, PACKET_SIZE) != PACKET_SIZE ) {
							if ( usr->bnc == bnc ) {
								killed = 1;
								}
							kill_bouncer(usr->bnc);
							}
						}

					if ( killed == 0 ) {
						*(packet + 2) = REQ_PASV;
						memcpy(packet + 3, pasv_value, 6);
						memcpy(packet + 9, &usr->uid, 2);
						if ( write(bnc->sock, packet, PACKET_SIZE) != PACKET_SIZE ) {
							kill_bouncer(usr->bnc);
							event = EVENT_FAIL_NOBNC;
							} else {
							usr->bnc = bnc;
							event = EVENT_HIDE;
							}
						} else {
						event = EVENT_FAIL_NOBNC;
						}
					} else {
					event = EVENT_FAIL_NOBNC;
					}
				}
			}

		switch ( event ) {
			case EVENT_FAIL_NOBNC:
				if ( add_to_s_buf(usr, "520 No active trafficbouncers.\r\n", 32) == -1 ) {
					return -1;
					}
				real_end_p += 32;
			case EVENT_HIDE:
				if ( p + 1 < real_end_p ) {
					memmove(line_p, p + 1, real_end_p - p);
					}
				usr->read_s_len -= real_len;
				break;
			case EVENT_OK:
				usr->write_u_len += real_len;
				break;
			}

		line_p = p + 1;
		newlines++;
		}
	}


	/***** Check if line is too long for us to buffer -> flush to user *****/

 if ( newlines == 0 ) {
	if ( usr->read_s_ignore == 1 ) {
		usr->write_u_len = usr->read_s_len;
		} else if ( usr->write_u_len == 0 && usr->read_s_len >= BUF_SIZE ) {
		usr->write_u_len = usr->read_s_len;
		usr->read_s_ignore = 1;
		}
	}
 return 0;
}











int main(int argc, char **argv) {
 fd_set			read_ds, write_ds;
 struct sockaddr_in	user_addr;
 struct sockaddr_in	bnc_addr;
 struct sockaddr_in	tmp_addr;
 struct _TB		*new_bnc;
 struct _TB		*bnc;
 struct _USER		*new_user;
 struct _USER		*usr;
 struct timeval		tv;
 struct timeval		*tv_p;
 unsigned short		p1, p2;
 unsigned short		t_uid;
 unsigned char		t_port[2];
 unsigned char		t_ip[30];
 unsigned char		t_str[512];
 time_t			until;
 char			*p;
 int			user_sock;
 int			bnc_sock;
 int			b_fd, g_fd;
 int			s, t;


 printf("dTbentry - %s\n", copyright);

	/****** Check arguments ******/

 if ( argc != 2 ) {
	printf("Usage: dtbentry <config>\n");
	exit(1);
	}

	/****** Null data ******/

 user = (struct _USER *)malloc(sizeof(struct _USER));
 tb = (struct _TB *)malloc(sizeof(struct _TB));

 bzero(&config, sizeof(struct _CONFIG));
 bzero(tb, sizeof(struct _TB));
 bzero(user, sizeof(struct _USER));
 bzero(&user_addr, sz);
 bzero(&bnc_addr, sz);
 last_bnc = tb;
 last_user = user;

 *packet = 2;
 *(packet + 1) = 0;


	/****** Drop evil signals ******/  

 signal(SIGPIPE, SIG_IGN);

	/****** Initialize random ******/

 gettimeofday(&tv, NULL);
 srandom(tv.tv_usec);

	/****** Read config ******/

#ifdef BLOWFISH
 printf("Password: ");
 sprintf(t_str, "%%%us", sizeof(config_key));
 scanf(t_str, config_key);
 config_key_len = strlen(config_key);
#endif

 if ( open_config(argv[1]) == -1 ) {
	printf("FATAL: Could not open config file!\n");
	exit(2);
	}

 rewind_config();
 s = get_config();
 close_config();

 if ( s != 0 ) {
	exit(2);
	}

 user_addr.sin_family = AF_INET;
 bnc_addr.sin_family = AF_INET;

	/****** Bind ports ******/

 user_addr.sin_port = htons(config.user_port);
 bnc_addr.sin_port = htons(config.bnc_port);

 user_sock = socket(AF_INET, SOCK_STREAM, 0);
 bnc_sock = socket(AF_INET, SOCK_STREAM, 0);

 if ( bind(user_sock, (struct sockaddr *)&user_addr, sz) != 0 ) {
	printf("FATAL: Could not bind listen socket to user port!\n");
	exit(2);
	}

 if ( bind(bnc_sock, (struct sockaddr *)&bnc_addr, sz) != 0 ) {
	printf("FATAL: Could not bind listen socket to bouncer port!\n");
	exit(2);
	}

 g_fd = (user_sock > bnc_sock ? user_sock : bnc_sock);

	/****** TLS ******/

#ifdef TLS
 tls_init();
#endif

	/****** FORK ******/

 if ( fork() ) {
	close(user_sock);
	close(bnc_sock);
	return -1;
	}

	/****** Listen ******/

 fcntl(user_sock, F_SETFL, O_NONBLOCK);
 fcntl(bnc_sock, F_SETFL, O_NONBLOCK);

 listen(user_sock, 128);
 listen(bnc_sock, 128);

 while ( 1 ) {
	FD_ZERO(&read_ds);
	FD_ZERO(&write_ds);

		/****** Listen for new connections ******/

	FD_SET(user_sock, &read_ds);
	FD_SET(bnc_sock, &read_ds);

	b_fd = g_fd;

	next_event = -1;

		/****** Listen for active user sockets ******/

	for ( usr = user->next ; usr != NULL ; usr = usr->next ) {
		switch ( usr->status ) {
				case U_WAIT_IDENT_CONNECT:
					s = usr->ident_sock;
					if (usr->timeout <= now) {
						close(s);

						s = socket(AF_INET, SOCK_STREAM, 0);
						fcntl(s, F_SETFL, O_NONBLOCK);

						usr->site_sock = s;
						usr->status = U_WAIT_CONNECT;

						usr->timeout = now + config.connect_timeout;

						connect(s, (struct sockaddr *)&config.site_addr, sz);
						}
					add_timer(usr->timeout);
					FD_SET(s, &write_ds);
					break;

				case U_WAIT_IDENT:
					s = usr->ident_sock;
					if (usr->timeout <= now) {
						close(s);

						s = socket(AF_INET, SOCK_STREAM, 0);
						fcntl(s, F_SETFL, O_NONBLOCK);

						usr->site_sock = s;
						usr->status = U_WAIT_CONNECT;

						usr->timeout = now + config.connect_timeout;

						connect(s, (struct sockaddr *)&config.site_addr, sz);
						FD_SET(s, &write_ds);
						} else {
						FD_SET(s, &read_ds);
						}
					add_timer(usr->timeout);
					break;

				case U_WAIT_CONNECT:
					if (usr->timeout <= now) {
						usr = kill_user(usr);
						continue;
						}
					add_timer(usr->timeout);
					FD_SET(s = usr->site_sock, &write_ds);
					break;
#ifdef TLS
				case U_WAIT_SSL_ACCEPT:
					switch ( usr->ssl_status ) {
						case S_WANT_READ:
							FD_SET(s = usr->user_sock, &read_ds);
							break;
						case S_WANT_WRITE:
							FD_SET(s = usr->user_sock, &write_ds);
							break;
						}
					break;
#endif

				case U_ACTIVE:
					s = usr->user_sock;
					t = usr->site_sock;

					if ( usr->read_u_len < BUF_SIZE ) {
						FD_SET(s, &read_ds);
						}
					if ( usr->read_s_len < BUF_SIZE ) {
						FD_SET(t, &read_ds);
						}
					if ( usr->write_u_len > 0 ) {
						FD_SET(s, &write_ds);
						}
					if ( usr->write_s_len > 0 ) {
						FD_SET(t, &write_ds);
						}

					s = (s > t ? s : t);
					break;
			}
		b_fd = (b_fd > s ? b_fd : s);
		}

	if ( next_event > now ) {
		tv.tv_sec = next_event - now;
		tv.tv_usec = 0;
		tv_p = &tv;
		} else {
		tv_p = 0;
		}

		/******* Listen for active bouncers ******/

	for ( bnc = tb->next ; bnc != NULL ; bnc = bnc->next ) {
		s = bnc->sock;
		FD_SET(s, &read_ds);
		b_fd = (b_fd > s ? b_fd : s);
		}

		/****** Wait until data arrives ******/

	s = select(b_fd + 1, &read_ds, &write_ds, 0, tv_p);

	now = time(NULL);

		/****** Start from beginning ******/

	if ( s == 0 ) {
		continue;
		}

		/****** Let new users in ******/

	if ( FD_ISSET(user_sock, &read_ds) ) {
		if ((s = accept(user_sock, (struct sockaddr *)&tmp_addr, &sz)) != -1) {
			new_user = (struct _USER *)malloc(sizeof(struct _USER));
			bzero(new_user, sizeof(struct _USER));
			memcpy(&new_user->addr, &tmp_addr, sz);
			new_user->user_sock = s;
			new_user->uid = s;
			new_user->u_buf = malloc(new_user->u_buf_size = BUF_SIZE);
			new_user->s_buf = malloc(new_user->s_buf_size = BUF_SIZE);

			s = socket(AF_INET, SOCK_STREAM, 0);
			fcntl(s, F_SETFL, O_NONBLOCK);

			if ( config.read_ident == 1 ) {
				new_user->status = U_WAIT_IDENT_CONNECT;
				new_user->ident_sock = s;
				new_user->timeout = now + config.ident_timeout;
				sprintf(new_user->ident, DEFAULT_IDENT);

				tmp_addr.sin_port = htons(IDENTD_PORT);
				connect(s, (struct sockaddr *)&tmp_addr, sz);
				} else {
				new_user->status = U_WAIT_CONNECT;
				new_user->site_sock = s;
				new_user->timeout = now + config.connect_timeout;
				connect(s, (struct sockaddr *)&config.site_addr, sz);
				}

			last_user->next = new_user;
			new_user->prev = last_user;
			last_user = new_user;
			}
		}

		/******* Let new bouncers in ********/

	if ( FD_ISSET(bnc_sock, &read_ds) ) {
		if ((s = accept(bnc_sock, (struct sockaddr *)&tmp_addr, &sz)) != -1) {
			new_bnc = (struct _TB *)malloc(sizeof(struct _TB));
			bzero(new_bnc, sizeof(struct _TB));

			new_bnc->sock = s;
			new_bnc->status = B_WAIT_AUTH;
			memcpy(&new_bnc->addr, &tmp_addr, sz);

			*(packet + 2) = REQ_AUTH;
			write(s, packet, PACKET_SIZE);

			last_bnc->next = new_bnc;
			new_bnc->prev = last_bnc;
			last_bnc = new_bnc;
			}
		}
		
		/******  Read what user sockets have to say ******/

	for ( usr = user->next ; usr != NULL ; usr = usr->next ) {
		switch ( usr->status ) {
				case U_WAIT_IDENT_CONNECT:
					if ( FD_ISSET(usr->ident_sock, &write_ds) ) {
						t = sizeof(s);
						getsockopt(usr->ident_sock, SOL_SOCKET, SO_ERROR, &s, &t);
						
						if ( s == 0 ) {
							fcntl(usr->ident_sock, F_SETFL, 0);
							usr->status = U_WAIT_IDENT;
							t = sprintf(t_str, "%u,%u\r\n", ntohs(usr->addr.sin_port), config.user_port);
							if ( write(usr->ident_sock, t_str, t) == t ) {
								continue;
								}
							}

						close(usr->ident_sock);

						s = socket(AF_INET, SOCK_STREAM, 0);
						fcntl(s, F_SETFL, O_NONBLOCK);

						usr->site_sock = s;
						usr->status = U_WAIT_CONNECT;
						usr->timeout = now + config.connect_timeout;

						connect(s, (struct sockaddr *)&config.site_addr, sz);
						}
					break;

				case U_WAIT_IDENT:
					if ( FD_ISSET(usr->ident_sock, &read_ds) ) {
						s = read(usr->ident_sock, t_str, sizeof(t_str));
						if ( s > 0 && usr->read_u_len + s < BUF_SIZE ) {
							memcpy(usr->u_buf + usr->read_u_len, t_str, s);
							usr->read_u_len += s;
							if ((char *)strchr(usr->u_buf, '\n') == NULL ) {
								continue;
								}
							}

						close(usr->ident_sock);

						if ( sscanf(usr->u_buf, "%u , %u : USERID :%*[^:]:%255s", &p1, &p2, usr->ident) != 3 ||
						    p1 != ntohs(usr->addr.sin_port) ||
						    p2 != config.user_port ) {
							sprintf(usr->ident, DEFAULT_IDENT);
							} else {
							for ( p = usr->ident ; *p != '\r' && *p ; p++ );
							*p = 0;
							}

						s = socket(AF_INET, SOCK_STREAM, 0);
						fcntl(s, F_SETFL, O_NONBLOCK);

						usr->read_u_len = 0;
						usr->site_sock = s;
						usr->status = U_WAIT_CONNECT;
						usr->timeout = now + config.connect_timeout;

						connect(s, (struct sockaddr *)&config.site_addr, sz);
						}
					break;

				/****** Waiting for connection to site to be established ******/

				case U_WAIT_CONNECT:
					if ( FD_ISSET(usr->site_sock, &write_ds) ) {
						t = sizeof(s);
						getsockopt(usr->site_sock, SOL_SOCKET, SO_ERROR, &s, &t);

						if ( s != 0 ) {
							usr = kill_user(usr);
							break;
							}

						fcntl(usr->user_sock, F_SETFL, O_NONBLOCK);
						usr->status = U_ACTIVE;
						if ( config.read_ident == 1 ) {
							memcpy(t_ip, &usr->addr.sin_addr, 4);
							sprintf(t_str, "%u.%u.%u.%u", *t_ip, *(t_ip + 1), *(t_ip + 2), *(t_ip + 3));
							usr->read_u_len = usr->write_s_len = sprintf(usr->u_buf, "IDNT %s@%s:%s\r\n", usr->ident, t_str, t_str);
							usr->read_s_len = usr->write_u_len = sprintf(usr->s_buf, "220- dTbnc - %s\r\n", copyright);
							continue;
							}
						}
					break;
#ifdef TLS
				case U_WAIT_SSL_ACCEPT:
					if ( usr->write_u_len > 0 ) {
						if ( FD_ISSET(usr->user_sock, &write_ds) ) {
							switch (s = write(usr->user_sock, usr->s_buf, usr->write_u_len)) {
								case -1:
									if ( errno == EAGAIN ) {
										break;
										}
									usr = kill_user(usr);
									continue;
								default:
									usr->write_u_len -= s;
									if ((usr->read_s_len -= s) > 0 ) {
										memmove(usr->s_buf, usr->s_buf + s, usr->read_s_len);
										} else {
										usr->ssl_status = S_WANT_READ;
										}
									break;
								}
							}
						} else if ( FD_ISSET(usr->user_sock, &read_ds)) {
						if ((s = SSL_accept(usr->ssl)) <= 0) {
							switch (SSL_get_error(usr->ssl, s) ) {
								case SSL_ERROR_WANT_READ:
								case SSL_ERROR_WANT_WRITE:
									usr->status = U_ACTIVE;
									break;
								case SSL_ERROR_ZERO_RETURN:
									usr->status = U_ACTIVE;
									SSL_free(usr->ssl);
									usr->ssl = NULL;
									break;
								default:
									usr = kill_user(usr);
									break;
								}
							} else {
							usr->status = U_ACTIVE;
							}
						}
					break;
#endif

				/****** Reading data from user sockets ******/

				case U_ACTIVE:

				/****** Read data from user if available ******/

					if ( FD_ISSET(usr->user_sock, &read_ds) ) {
						switch (s = READ(usr->ssl, usr->user_sock, usr->u_buf + usr->read_u_len, BUF_SIZE - usr->read_u_len)) {
							case -1:
								if ( errno == EAGAIN ) {
									break;
									}
							case 0:
								usr = kill_user(usr);
								continue;
							default:
								if ( manage_new_u_data(usr, s) == -1 ) {
									usr = kill_user(usr);
									}
								break;
							}
						}

				/****** Read data from site if available *******/

					if ( FD_ISSET(usr->site_sock, &read_ds) ) {
						switch (s = read(usr->site_sock, usr->s_buf + usr->read_s_len, BUF_SIZE - usr->read_s_len)) {
							case -1:
								if ( errno == EAGAIN ) {
									break;
									}
							case 0:
								usr = kill_user(usr);
								continue;
							default:
								if ( manage_new_s_data(usr, s) == -1 ) {
									usr = kill_user(usr);
									}
								break;
							}
						}

				/****** Write data to user if available and accepted ******/

					if ( FD_ISSET(usr->user_sock, &write_ds) ) {
						switch (s = WRITE(usr->ssl, usr->user_sock, usr->s_buf, usr->write_u_len)) {
							case -1:
								if ( errno == EAGAIN ) {
									break;
									}
								usr = kill_user(usr);
								continue;
							default:
								usr->write_u_len -= s;
								if ((usr->read_s_len -= s) > 0 ) {
									memmove(usr->s_buf, usr->s_buf + s, usr->read_s_len);
									}
								break;
							}
						}

				/****** Write data to site if available and accepted ******/

					if ( FD_ISSET(usr->site_sock, &write_ds) ) {
						switch (s = write(usr->site_sock, usr->u_buf, usr->write_s_len)) {
							case -1:
								if ( errno == EAGAIN ) {
									break;
									}
								usr = kill_user(usr);
								continue;
							default:
								usr->write_s_len -= s;
								if ((usr->read_u_len -= s) > 0 ) {
									memmove(usr->u_buf, usr->u_buf + s, usr->read_u_len);
									}
								break;
							}
						}
					break;
			}
		}

		/****** Read what traffic bouncers have to say ******/


	for ( bnc = tb->next ; bnc != NULL ; bnc = bnc->next ) {
		if ( FD_ISSET(bnc->sock, &read_ds)) {
			s = read(bnc->sock, bnc->packet + bnc->packet_len, PACKET_SIZE - bnc->packet_len);
			if (((bnc->packet_len += s) == PACKET_SIZE && (*bnc->packet != 2 || *(bnc->packet + 1) != 0)) || s <= 0 ) {
				bnc = kill_bouncer(bnc);
				} else if ( bnc->packet_len == PACKET_SIZE ) {
				bnc->packet_len = 0;

				if ( bnc->status == B_WAIT_AUTH && *(bnc->packet + 2) != AUTH_REPLY ) {
					bnc = kill_bouncer(bnc);
					continue;
					}

				switch (*(bnc->packet + 2)) {

					/****** Response to REQ_AUTH ******/

					case AUTH_REPLY:
						if ( *(bnc->packet + 3) == config.key_len &&
						    ! memcmp(bnc->packet + 4, config.key, config.key_len)) {
							bnc->status = B_ACTIVE;
							} else {
							bnc = kill_bouncer(bnc);
							}
						break;

					/****** Response to REQ_PASV ******/

					case PASV_REPLY:
						memcpy(&t_uid, bnc->packet + 4, 2);
						for ( usr = user->next ; usr != NULL ; usr = usr->next ) {
							if ( usr->uid == t_uid ) {
								if ( *(bnc->packet + 3) == 0 ) {
									memcpy(t_ip, bnc->packet + 6, 4);
									memcpy(t_port, bnc->packet + 10, 2);
									s = sprintf(t_str, "227 Entering Passive Mode (%hu,%hu,%hu,%hu,%u,%u)\r\n", *t_ip, *(t_ip + 1), *(t_ip + 2), *(t_ip + 3), *t_port, *(t_port + 1));
									usr->data_status = D_ACTIVE;
									if ( add_to_s_buf(usr, t_str, s) == -1 ) {
										usr = kill_user(usr);
										}
									} else {
									usr->data_status = D_INACTIVE;
									if ( add_to_s_buf(usr, "520 PASV command failed.\r\n", 26) == -1 ) {
										usr = kill_user(usr);
										}
									}
								break;
								}
							}
						break;

					/****** Response to REQ_PORT ******/

					case PORT_REPLY:
						memcpy(&t_uid, bnc->packet + 4, 2);
						for ( usr = user->next ; usr != NULL ; usr = usr->next ) {
							if ( usr->uid == t_uid ) {
								if ( *(bnc->packet + 3) == 0 ) {
									memcpy(t_ip, bnc->packet + 6, 4);
									memcpy(t_port, bnc->packet + 10, 2);
									s = sprintf(t_str, "PORT %hu,%hu,%hu,%hu,%hu,%hu\r\n", *t_ip, *(t_ip + 1), *(t_ip + 2), *(t_ip + 3), *t_port, *(t_port + 1));
									usr->data_status = D_ACTIVE;
									if ( add_to_u_buf(usr, t_str, s) == - 1 ) {
										usr = kill_user(usr);
										}
									} else {
									usr->data_status = D_INACTIVE;
									if ( add_to_s_buf(usr, "520 PORT command failed.\r\n", 26) == -1 ) {
										usr = kill_user(usr);
										}
									}
								break;
								}
							}
						break;

					/****** Stats response ******/

					case SET_STAT:
						bnc->status = *(bnc->packet + 3);
						memcpy(&bnc->kbytes_total, bnc->packet + 4, 4);
						memcpy(&bnc->kbytes_left, bnc->packet + 8, 4);
						memcpy(&bnc->time_total, bnc->packet + 12, 4);
						memcpy(&bnc->kbytes_xfer, bnc->packet + 16, 4);
						memcpy(&bnc->next_reset, bnc->packet + 20, 4);
						memcpy(&bnc->transfers_active, bnc->packet + 24, 2);
						bnc->next_reset += now;
						break;

					/****** Ping - send Pong ******/

					case REQ_PING:
						*(packet + 2) = PONG_REPLY;
						write(bnc->sock, packet, PACKET_SIZE);
						break;
					}
				}
			}
		}
	}
}
