#define PACKET_SIZE		35
#define MAX_ENTRIES		10
#define BUF_SIZE		65535
#define	DEBUG			0

#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/time.h>
#include <netdb.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <resolv.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <resolv.h>
#include <stdarg.h>
#include <signal.h>
#include <errno.h>
#include "global.h"
#include "config-reader.h"

static char copyright[] = "Copyright (c) dark0n3 2001 - 2002 <darkone@glftpd.com>";

struct _ENTRY {
	struct sockaddr_in	addr;
	struct sockaddr_in	bind_client_addr;
	struct sockaddr_in	bind_server_addr;
	unsigned char		*key;
	unsigned char		key_len;
	unsigned char		status;
	unsigned char		ping;
	unsigned char		packet[PACKET_SIZE];
	unsigned char		packet_len;
	unsigned char		client_is_bind;
	unsigned char		server_is_bind;
	int			sock;
};

struct _STAT {
	unsigned long long	total_bytes;
	unsigned long long	left_bytes;
	unsigned long		total_time;
	unsigned short		transfers;
	unsigned char		status;
	time_t			transfer_start;
};

struct _CONF {
	unsigned long long	transfer_limit;
	unsigned long 		reset_interval;
	unsigned long		connect_interval;
	unsigned long		ping_interval;
	unsigned long		ping_timeout;
	unsigned long		connect_timeout;
	unsigned long		transfer_timeout;
	unsigned short		port_min;
	unsigned short		port_max;
	unsigned short		port_last;
	unsigned char		port_random;
	unsigned char		entries;
};

struct _XFER {
	unsigned char		buffer[BUF_SIZE];
	unsigned short		buf_size;
	struct sockaddr_in	conn_addr;
	struct sockaddr_in	listen_addr;
	unsigned short		uid;
	unsigned char		type;
	int			entry_sock;
	int			conn_sock;
	int			listen_sock;
	int			read_sock;
	int			write_sock;
	int			entry;
	unsigned char		status;
	time_t			timeout;
	struct _XFER		*next;
	struct _XFER		*prev;
};

struct _TIMER {
	time_t			reset;
	time_t			ping;
	time_t			ping_timeout;
	time_t			connect;
	time_t			connect_timeout;
};

#ifdef BLOWFISH
 char		config_key[256];
 int		config_key_len;
#endif

struct _CONF	config;
struct _STAT	stats;
struct _ENTRY	entry[MAX_ENTRIES];
struct _TIMER	timer;
struct _XFER	*transfer;
struct _XFER	*last_xfer;

unsigned char	packet[PACKET_SIZE];
time_t		now;
time_t		next_event = 0;
int		sz = sizeof(struct sockaddr_in);
extern int	errno;


#if ( DEBUG )
#define	D_LOG	log
#else
#define D_LOG
#endif

#define add_timer(t)	if (t < next_event) t = next_event
#define set_timer(t)	next_event = t



/*
 *
 * Appends stuff to LOGFILE
 *
 */
void log(char *fmt, ...) {
 time_t         timenow;
 FILE           *file; 
 va_list        ap;  

 va_start(ap, fmt);
 timenow = time(NULL);

 if ((file = fopen("debug", "a+")) != NULL ) {
	fprintf(file, "%.24s - ", ctime(&timenow));
	vfprintf(file, fmt, ap);
	fclose(file);
	}
 va_end(ap);
}




/*
 * 
 * Sends stats to active entries
 *
 * Todo: Add entry killing, if write fails
 *
 */
int send_stats(int which) {
 unsigned long		t_ulong;
 unsigned short		t_u;
 int			n;

 D_LOG("Sending stats\n");

 *(packet + 2) = SET_STAT;
 *(packet + 3) = stats.status;

 t_ulong = stats.total_bytes >> 10;
 memcpy(packet + 4, &t_ulong, 4);

 t_ulong = stats.left_bytes >> 10;
 memcpy(packet + 8, &t_ulong, 4);

 t_ulong = stats.total_time;
 if ( stats.transfers > 0 ) {
	t_ulong += now - stats.transfer_start;
	}
 if ( t_ulong == 0 ) {
	t_ulong = 1;
	}
 memcpy(packet + 12, &t_ulong, 4);

 t_ulong = (config.transfer_limit << 10) - (stats.left_bytes >> 10);
 memcpy(packet + 16, &t_ulong, 4);

 t_ulong = timer.reset - now;
 memcpy(packet + 20, &t_ulong, 4);

 t_u = stats.transfers;
 memcpy(packet + 24, &t_u, 2);
 
 if ( which == -1 ) {
	for ( n = 0 ; n < config.entries ; n++ ) {
		if ( entry[n].status == E_ACTIVE ) {
			write(entry[n].sock, packet, PACKET_SIZE);
			}
		}
	} else {
	write(entry[which].sock, packet, PACKET_SIZE);
	}

 D_LOG("Stats sent\n");
}




/*
 *
 * Frees memory and sockets for transfer
 *
 */
struct _XFER *destroy_xfer(struct _XFER *xfer) {
 struct _XFER	*prev_xfer;

 stats.transfers--;

 if ( stats.transfers == 0 ) {
	stats.total_time += now - stats.transfer_start;
	}

 if ( xfer->conn_sock > 0 ) {
	close(xfer->conn_sock);
	}
 if ( xfer->listen_sock > 0 ) {
	close(xfer->listen_sock);
	}

 prev_xfer = xfer->prev;
 if ( xfer->next ) {
	prev_xfer->next = xfer->next;
	(xfer->next)->prev = prev_xfer;
	} else {
	prev_xfer->next = NULL;
	last_xfer = prev_xfer;
	}

 free(xfer);

 return prev_xfer;
}





/*
 *
 * Kills transfer and updates stats
 *
 */
struct _XFER *kill_transfer(struct _XFER *xfer) {
 struct _XFER	*prev_xfer;

 D_LOG("Destroying transfer\n");

 prev_xfer = destroy_xfer(xfer);
 send_stats(-1);

 D_LOG("Transfer destroyed\n");

 return prev_xfer;
}






/*
 *
 * Closes all sockets related to this entry bnc
 *
 */
void kill_entry(struct _ENTRY *bnc) {
 struct	_XFER	*xfer;
 int		n, s;

 D_LOG("Destroying entry\n");

	/****** Wipe all non active transfers ******/

 if ((s = bnc->sock) > 0 ) {
	close(s);
	for ( xfer = transfer->next ; xfer != NULL ; xfer = xfer->next ) {
		if ( xfer->entry_sock == s && xfer->status < U_ACTIVE ) {
			xfer = destroy_xfer(xfer);
			}
		}
	}

	/****** Set status ******/

 bnc->status = E_OFFLINE;
 bnc->ping = 0;
 bnc->sock = 0;
 bnc->packet_len = 0;

 send_stats(-1);

 D_LOG("Entry destroyed\n");
}




/*
 *
 * Reads config file
 *
 */
int get_config() {
 struct hostent		*hp;
 unsigned char		len;
 char			*key;
 char			*value;

 while ((key = seek_config_next_branch()) != NULL ) {
	if ( ! memcmp(key, "global", 7)) {

		if ((value = seek_config_key("random ports")) == NULL) {
			printf("FATAL: 'random ports' was not properly defined in global!\n");
			return -1;
			}
		if ( ! strcasecmp(value, "true")) {
			config.port_random = 1;
			} else if ( ! strcasecmp(value, "false")) {
			config.port_random = 0;
			} else {
			printf("FATAL: 'random ports' was not properly defined in global!\n");
			return -1;
			}

		if ((value = seek_config_key("port min")) == NULL ||
		    (config.port_min = atoi(value)) == 0 ) {
			printf("FATAL: 'port min' was not properly defined in global!\n");
			return -1;
			}

		if ((value = seek_config_key("port max")) == NULL ||
		    (config.port_max = atoi(value)) == 0 ) {
			printf("FATAL: 'port max' was not properly defined in global!\n");
			return -1;
			}

		if ((value = seek_config_key("connect interval")) == NULL ||
		    (config.connect_interval = atoi(value)) == 0 ) {
			printf("FATAL: 'connect interval' was not properly defined in global!\n");
			return -1;
			}

		if ((value = seek_config_key("connect timeout")) == NULL ||
		    (config.connect_timeout = atoi(value)) == 0 ) {
			printf("FATAL: 'connect timeout' was not properly defined in global!\n");
			return -1;
			}

		if ((value = seek_config_key("transfer timeout")) == NULL ||
		    (config.transfer_timeout = atoi(value)) == 0 ) {
			printf("FATAL: 'transfer timeout' was not properly defined in global!\n");
			return -1;
			}

		if ((value = seek_config_key("ping interval")) == NULL ||
		    (config.ping_interval = atoi(value)) == 0 ) {
			printf("FATAL: 'ping interval' was not properly defined in global!\n");
			return -1;
			}

		if ((value = seek_config_key("ping timeout")) == NULL ||
		    (config.ping_timeout = atoi(value)) == 0 ) {
			printf("FATAL: 'ping timeout' was not properly defined in global!\n");
			return -1;
			}

		if ((value = seek_config_key("transfer limit")) == NULL ||
		    (config.transfer_limit = atoi(value)) == 0 ) {
			printf("FATAL: 'transfer limit' was not properly defined in global!\n");
			return -1;
			}

		if ((value = seek_config_key("reset interval")) == NULL ||
		    (config.reset_interval = atoi(value)) == 0 ) {
			printf("FATAL: 'reset interval' was not properly defined in global!\n");
			return -1;
			}

		} else {
		if ((value = seek_config_key("entry host")) == NULL ||
		    (hp = gethostbyname(value)) == NULL ) {
			continue;
			}
		memcpy(&entry[config.entries].addr.sin_addr.s_addr, hp->h_addr, hp->h_length);

		if ((value = seek_config_key("entry port")) == NULL ||
		    (entry[config.entries].addr.sin_port = htons(atoi(value))) == 0 ) {
			continue;
			}

		if ((value = seek_config_key("bind client host")) != NULL ) {
			if ((hp = gethostbyname(value)) == NULL ) {
				return -1;
				}
			memcpy(&entry[config.entries].bind_client_addr.sin_addr.s_addr, hp->h_addr, hp->h_length);
			entry[config.entries].bind_client_addr.sin_family = AF_INET;
			entry[config.entries].client_is_bind = 1;
			}

		if ((value = seek_config_key("bind server host")) != NULL ) {
			if ((hp = gethostbyname(value)) == NULL ) {
				return -1;
				}
			memcpy(&entry[config.entries].bind_server_addr.sin_addr.s_addr, hp->h_addr, hp->h_length);
			entry[config.entries].bind_server_addr.sin_family = AF_INET;
			entry[config.entries].server_is_bind = 1;
			}

		if ((value = seek_config_key("entry pass")) == NULL ||
		    (len = strlen(value)) == 0 ) {
			continue;
			}

		entry[config.entries].addr.sin_family = AF_INET;

		len = (len < PACKET_SIZE - 4 ? len : PACKET_SIZE - 4);
		entry[config.entries].key = malloc(len);
		memcpy(entry[config.entries].key, value, len);
		entry[config.entries].key_len = len;
		config.entries++;
		}
	}
 return 0;
}







int main(int argc, char **argv) {
 struct timeval		tv_timer;
 struct _XFER		*xfer, *new_xfer;
 unsigned short		t_u;
 fd_set			read_ds;
 fd_set			write_ds;
 char			t_str[10];
 int			changes = 0;
 int			n, s, t;
 int			b_fd = 0;
 int			u_fd_max = 0;
 int			e_fd_max = 0;

 printf("dTbnc - %s\n", copyright);

	/****** Check arguments ******/

 if ( argc != 2 ) {
	printf("Usage: dtbnc <config>\n");
	exit(1);
	}


	/****** Clear structures ******/

 transfer = (struct _XFER *)malloc(sizeof(struct _XFER));

 bzero(&config, sizeof(struct _CONF));
 bzero(transfer, sizeof(struct _XFER));
 bzero(&entry, sizeof(struct _ENTRY) * MAX_ENTRIES);
 bzero(&stats, sizeof(struct _STAT));
 bzero(&timer, sizeof(struct _TIMER));
 stats.status = B_ACTIVE;
 last_xfer = transfer;

 *packet = 2;
 *(packet + 1) = 0;

	/****** Drop evil signals ******/

 signal(SIGPIPE, SIG_IGN);

	/****** Initialize random ******/

 gettimeofday(&tv_timer,NULL);
 srandom(tv_timer.tv_usec);

	/****** Read config ******/

#ifdef BLOWFISH
 printf("Password: ");
 sprintf(t_str, "%%%us", sizeof(config_key));
 scanf(t_str, config_key);
 config_key_len = strlen(config_key);
#endif

 if ( open_config(argv[1]) == -1 ) {
	printf("FATAL: Could not open config file!\n");
	exit(2);
	}

 rewind_config();
 s = get_config();
 close_config();

 if ( s != 0 ) {
	exit(2);
	}

 now = time(NULL);

 timer.reset = now + config.reset_interval;
 stats.left_bytes = config.transfer_limit << 20;
 config.port_last = config.port_max;

	/****** FORK ******/

 if ( fork() ) {
	return -1;
	}

	/****** Infinite loop ******/

 while ( 1 ) {
	FD_ZERO(&read_ds);
	FD_ZERO(&write_ds);

		/****** Event timer ******/

	if ( now >= next_event ) {

		D_LOG("TIMER 1\n");

			/****** Check for connect ******/

		if (timer.connect <= now ) {
			for ( n = 0 ; n < config.entries; n++ ) {
				if ( entry[n].status == E_OFFLINE && (s = socket(AF_INET, SOCK_STREAM, 0)) > 0 ) {
					if ( entry[n].client_is_bind ) {
						entry[n].bind_client_addr.sin_port = 0;
						bind(s, (struct sockaddr *)&entry[n].bind_client_addr, sz);
						}
					fcntl(s, F_SETFL, O_NONBLOCK);
					e_fd_max = (s > e_fd_max ? s : e_fd_max);
					entry[n].sock = s;
					entry[n].status = E_WAIT_CONNECT;
					connect(s, (struct sockaddr *)&entry[n].addr, sz);
					}
				}
			timer.connect = now + config.connect_interval;
			timer.connect_timeout = now + config.connect_timeout;
			}
		set_timer(timer.connect);

		D_LOG("TIMER 2\n");

			/****** Check for connect timeout ******/

		if ( timer.connect_timeout > 0 ) {
			if (timer.connect_timeout <= now ) {
				for ( n = 0 ; n < config.entries; n++ ) {
					if ( entry[n].status == E_WAIT_CONNECT ) {
						kill_entry(&entry[n]);
						}
					}
				timer.connect_timeout = 0;
				} else {
				add_timer(timer.connect_timeout);
				}
			}



		D_LOG("TIMER 3\n");

			/****** Check for ping ******/

		if (timer.ping <= now) {
			*(packet + 2) = REQ_PING;

			for ( n = 0 ; n < config.entries; n++ ) {
				if ( entry[n].status == E_ACTIVE ) {
					entry[n].ping = 1;
					write(entry[n].sock, packet, PACKET_SIZE);
					}
				}

			timer.ping = now + config.ping_interval;
			timer.ping_timeout = now + config.ping_timeout;
			}
		add_timer(timer.ping);

		D_LOG("TIMER 4\n");

			/****** Check for entries ping timeout ******/

		if ( timer.ping_timeout > 0 ) {
			if (timer.ping_timeout <= now) {
				for ( n = 0 ; n < config.entries ; n++ ) {
					if ( entry[n].status == E_ACTIVE && entry[n].ping == 1 ) {
						kill_entry(&entry[n]);
						}
					}
				timer.ping_timeout = 0;
				} else {
				add_timer(timer.ping_timeout);
				}
			}

		D_LOG("TIMER 5\n");

			/****** Check for stats reset ******/

		if (timer.reset <= now) {
			stats.status = B_ACTIVE;
			stats.left_bytes = config.transfer_limit << 20;

			timer.reset = now + config.reset_interval;

			send_stats(-1);
			}

		add_timer(timer.reset);

		D_LOG("TIMER 6\n");

			/****** Check for transfer timeouts ******/

		*(packet + 3) = 1;

		for ( xfer = transfer->next; xfer != NULL; xfer = xfer->next ) {
			switch ( xfer->status ) {
				case U_WAIT_CONNECT:
					if (xfer->timeout <= now) {
						*(packet + 2) = xfer->type;
						memcpy(packet + 4, &xfer->uid, 2);
						write(xfer->entry_sock, packet, PACKET_SIZE);
						xfer = kill_transfer(xfer);
						} else {
						add_timer(xfer->timeout);
						}
					break;

				case U_WAIT_ACCEPT:
					if (xfer->timeout <= now) {
						xfer = kill_transfer(xfer);
						} else {
						add_timer(xfer->timeout);
						}
					break;
				}
			}

			/****** Configure timer for next event ******/
		}

	tv_timer.tv_sec = next_event - now;
	tv_timer.tv_usec = 0;

	for ( xfer = transfer->next; xfer != NULL ; xfer = xfer->next ) {
		switch(xfer->status) {

			case U_WAIT_CONNECT:
				FD_SET(xfer->conn_sock, &write_ds);
				break;

			case U_WAIT_ACCEPT:
				FD_SET(xfer->listen_sock, &read_ds);
				break;

			case U_XFER_DIRECTION:
				FD_SET(xfer->listen_sock, &read_ds);
				FD_SET(xfer->conn_sock, &read_ds);
				break;

			case U_ACTIVE:
				switch ( xfer->buf_size ) {
					case BUF_SIZE:
						FD_SET(xfer->write_sock, &write_ds);
						break;
					case 0:
						FD_SET(xfer->read_sock, &read_ds);
						break;
					default:
						FD_SET(xfer->read_sock, &read_ds);
						FD_SET(xfer->write_sock, &write_ds);
						break;
					}
				break;

			case U_WAIT_END:
				FD_SET(xfer->write_sock, &write_ds);
				break;
			}
		}

	for (n = 0; n < config.entries; n++) {
		switch (entry[n].status) {
			case E_WAIT_CONNECT:
				FD_SET(entry[n].sock, &write_ds);
				break;
			case E_ACTIVE:
				FD_SET(entry[n].sock, &read_ds);
				break;
			}
		}

	b_fd = (u_fd_max > e_fd_max ? u_fd_max : e_fd_max) + 1;

	changes = select(b_fd, &read_ds, &write_ds, 0, &tv_timer);

	now = time(NULL);

	if ( changes == 0 ) {
		continue;
		}

	for ( n = 0 ; n < config.entries ; n++ ) {
		switch (entry[n].status) {

			case E_WAIT_CONNECT:
				if (FD_ISSET(entry[n].sock, &write_ds)) {
					t = sizeof(s);
					getsockopt(entry[n].sock, SOL_SOCKET, SO_ERROR, &s, &t);
					if ( s == 0 ) {
						fcntl(entry[n].sock, F_SETFL, 0);
						entry[n].status = E_ACTIVE;
						} else {
						kill_entry(&entry[n]);
						}
					}
				break;

			case E_ACTIVE:
				if (FD_ISSET(entry[n].sock, &read_ds)) {
					s = read(entry[n].sock, entry[n].packet + entry[n].packet_len, PACKET_SIZE - entry[n].packet_len);
					if (((entry[n].packet_len += s) == PACKET_SIZE && (*entry[n].packet != 2 || *(entry[n].packet + 1) != 0)) || s <= 0 ) {
						kill_entry(&entry[n]);
						} else if ( entry[n].packet_len == PACKET_SIZE ) {
						entry[n].packet_len = 0;
						switch (*(entry[n].packet + 2)) {
							case REQ_PASV:
							case REQ_PORT:
								D_LOG("REQ_PORT - BEGIN\n");
								if ((s = socket(AF_INET, SOCK_STREAM, 0)) >= 0) {
									new_xfer = (struct _XFER *)malloc(sizeof(struct _XFER));
									bzero(new_xfer, sizeof(struct _XFER));

									u_fd_max = (s > u_fd_max ? s : u_fd_max);
									fcntl(s, F_SETFL, O_NONBLOCK);

									memcpy(&new_xfer->uid, entry[n].packet + 9, 2);
									memcpy(&new_xfer->conn_addr.sin_addr.s_addr, entry[n].packet + 3, 4);
									memcpy(&new_xfer->conn_addr.sin_port, entry[n].packet + 7, 2);

									if ( *(entry[n].packet + 2) == REQ_PORT ) {
										new_xfer->type = PORT_REPLY;
										if ( entry[n].client_is_bind ) {
											entry[n].bind_client_addr.sin_port = 0;
											bind(s, (struct sockaddr *)&entry[n].bind_client_addr, sz);
											}
										} else {
										new_xfer->type = PASV_REPLY;
										if ( entry[n].server_is_bind ) {
											entry[n].bind_server_addr.sin_port = 0;
											bind(s, (struct sockaddr *)&entry[n].bind_server_addr, sz);
											}
										}

									new_xfer->conn_addr.sin_family = AF_INET;
									new_xfer->conn_sock = s;
									new_xfer->status = U_WAIT_CONNECT;
									new_xfer->timeout = now + config.transfer_timeout;
									new_xfer->entry_sock = entry[n].sock;
									new_xfer->entry = n;

									connect(s, (struct sockaddr *)&new_xfer->conn_addr, sz);
									add_timer(new_xfer->timeout);

									stats.transfers++;
									if ( stats.transfers == 1 ) {
										stats.transfer_start = now;
										}
									last_xfer->next = new_xfer;
									new_xfer->prev = last_xfer;
									last_xfer = new_xfer;
									send_stats(-1);
									} else {
									*(packet + 2) = *(entry[n].packet + 2);
									*(packet + 3) = 1;
									write(entry[n].sock, packet, PACKET_SIZE);
									}
								D_LOG("REQ_PORT - END\n");
								break;

							case REQ_CLEAN:
								D_LOG("REQ_CLEAN - BEGIN\n");
								memcpy(&t_u, entry[n].packet + 3, 2);
								s = entry[n].sock;
								for ( xfer = transfer->next ; xfer != NULL ; xfer = xfer->next ) {
									if ( t_u == xfer->uid && s == xfer->entry_sock ) {
										xfer = kill_transfer(xfer);
										}
									}
								D_LOG("REQ_CLEAN - END\n");
								break;

							case REQ_AUTH:
								*(packet + 2) = AUTH_REPLY;
								*(packet + 3) = entry[n].key_len;
								memcpy(packet + 4, entry[n].key, entry[n].key_len);
								write(entry[n].sock, packet, PACKET_SIZE);
								send_stats(n);
								break;

							case PONG_REPLY:
								entry[n].ping = 0;
								break;
							}
						}
					}
				break;
			}
		}

	/****** Listen for connectivity changes ******/

	for ( xfer = transfer->next ; xfer != NULL ; xfer = xfer->next ) {
		switch(xfer->status) {
			case U_WAIT_CONNECT:
				if (FD_ISSET(xfer->conn_sock, &write_ds)) {
					D_LOG("U_WAIT_CONNECT - BEGIN\n");
					t = sizeof(s);
					getsockopt(xfer->conn_sock, SOL_SOCKET, SO_ERROR, &s, &t);

					if((*(packet + 2) = xfer->type) == PORT_REPLY ) {
						memcpy(packet + 6, &entry[xfer->entry].bind_server_addr.sin_addr.s_addr, 4);
						} else {
						memcpy(packet + 6, &entry[xfer->entry].bind_client_addr.sin_addr.s_addr, 4);
						}

					memcpy(packet + 4, &xfer->uid, 2);

					if ( s == 0 && (s = socket(AF_INET, SOCK_STREAM, 0)) > 0) {
						xfer->status = U_WAIT_ACCEPT;

						u_fd_max = (s > u_fd_max ? s : u_fd_max);
						xfer->listen_addr.sin_family = AF_INET;

						t_u = config.port_last;
						do {
							if (config.port_random == 1) {
								t_u = ((config.port_max - config.port_min) * 1. * rand()/(RAND_MAX+1.0));
								t_u += config.port_min;
								} else {
								t_u++;
								if ( t_u > config.port_max ) t_u = config.port_min;
								}

							xfer->listen_addr.sin_port = htons(t_u);

							} while (bind(s, (struct sockaddr *)&xfer->listen_addr, sz) != 0);
						config.port_last = t_u;
						fcntl(s, F_SETFL, O_NONBLOCK);
						listen(s, 1);
						xfer->listen_sock = s;

						*(packet + 3) = 0;
						memcpy(packet + 10, &xfer->listen_addr.sin_port, 2);
						write(xfer->entry_sock, packet, PACKET_SIZE);

						} else {
						*(packet + 3) = 1;
						write(xfer->entry_sock, packet, PACKET_SIZE);
						xfer = kill_transfer(xfer);
						}
					D_LOG("U_WAIT_CONNECT - END\n");
					}
				break;

			case U_WAIT_ACCEPT:
				if (FD_ISSET(xfer->listen_sock, &read_ds)) {
					t = sz;
					if ((s = accept(xfer->listen_sock, (struct sockaddr *)&xfer->listen_addr, &t)) != -1) {
						D_LOG("U_WAIT_ACCEPT - BEGIN\n");
						close(xfer->listen_sock);
						xfer->listen_sock = s;
						fcntl(s, F_SETFL, O_NONBLOCK);
						u_fd_max = (s > u_fd_max ? s : u_fd_max);
						xfer->status = U_XFER_DIRECTION;
						D_LOG("U_WAIT_ACCEPT - END\n");
						}
					}
				break;

			case U_XFER_DIRECTION:
				if (FD_ISSET(xfer->listen_sock, &read_ds)) {
					xfer->read_sock = xfer->listen_sock;
					xfer->write_sock = xfer->conn_sock;
					} else if ( FD_ISSET(xfer->conn_sock, &read_ds)) {
					xfer->read_sock = xfer->conn_sock;
					xfer->write_sock = xfer->listen_sock;
					} else {
					break;
					}
				D_LOG("U_XFER_DIRECTION - BEGIN\n");
				t_u = BUF_SIZE;
				setsockopt(xfer->read_sock, SOL_SOCKET, SO_RCVBUF, &t_u, sizeof(unsigned short));
				setsockopt(xfer->write_sock, SOL_SOCKET, SO_SNDBUF, &t_u, sizeof(unsigned short));
				xfer->status = U_ACTIVE;
				D_LOG("U_XFER_DIRECTION - END\n");
				break;

			case U_ACTIVE:
				if (FD_ISSET(xfer->read_sock, &read_ds)) {
					switch(s = read(xfer->read_sock, xfer->buffer + xfer->buf_size, BUF_SIZE - xfer->buf_size)) {
						case -1:
							if ( errno == EAGAIN ) {
								break;
								}
						case 0:
							xfer->status = U_WAIT_END;
							break;
						default:
							xfer->buf_size += s;
							stats.total_bytes += s;

							if ( stats.left_bytes <= s ) {
								if ( stats.status == B_ACTIVE ) {
									stats.status = B_CLOSED;
									stats.left_bytes = 0;
									send_stats(-1);
									}
								} else {
								stats.left_bytes -= s;
								}
							break;
						}
					}

				if (xfer->buf_size > 0 && FD_ISSET(xfer->write_sock, &write_ds)) {
					switch(s = write(xfer->write_sock, xfer->buffer, xfer->buf_size)) {
						case -1:
							if ( errno == EAGAIN ) {
								break;
								}
							xfer = kill_transfer(xfer);
							break;
						default:
							if ((xfer->buf_size -= s) > 0) {
								memmove(xfer->buffer, xfer->buffer + s, xfer->buf_size);
								}
							break;
						}
					}
				break;

			case U_WAIT_END:
				if (FD_ISSET(xfer->write_sock, &write_ds)) {
					switch(s = write(xfer->write_sock, xfer->buffer, xfer->buf_size)) {
						case -1:
							if ( errno == EAGAIN ) {
								break;
								}
							xfer = kill_transfer(xfer);
							break;
						default:
							if ((xfer->buf_size -= s) == 0) {
								xfer = kill_transfer(xfer);
								} else {
								memmove(xfer->buffer, xfer->buffer + s, xfer->buf_size);
								}
							break;
						}
					}
				break;
			}
		}
	}
 D_LOG("Something got us here\n");
}
